"""
Module to extract and replace non-human names
"""
from preprocessing_pgp.name.const import REPLACE_HUMAN_REG_DICT


def replace_non_human_reg(name: str) -> str:
    """
    Replace non-human term with correct term using the dictionary based

    Parameters
    ----------
    name : str
        The name to preprocess replace non-human term

    Returns
    -------
    str
        The clean name containing the transformed non-human names
    """
    for word, to_word in REPLACE_HUMAN_REG_DICT.items():
        name = name.replace(word, to_word)
    return name.strip()



"""This module meant for preprocessing the card id before validation"""

import re
from typing import Tuple
from dataclasses import dataclass

import pandas as pd

from preprocessing_pgp.utils import sep_display


@dataclass
class CardIDCleaner:
    """
    Class contains cleansing functions for preprocessing card_id
    """

    def __remove_spaces(self, card_id: str) -> str:
        """
        Function to remove all spaces in card_id

        Parameters
        ----------
        card_id: str
            Input card_id to clean

        Returns
        -------
        str
            The output card_id without any spacing
        """

        # Remove spaces in between
        clean_card_id = re.sub(' +', '', card_id)
        clean_card_id = clean_card_id.strip()

        return clean_card_id

    def __remove_special_characters(self, card_id: str) -> str:
        """
        Removing special characters in card_id

        Parameters
        ----------
        card_id: str
            Input card_id to clean

        Returns
        -------
        str
            The output card_id without any punctuation
        """

        # translator = str.maketrans('', '', punctuation)

        # clean_card_id = card_id.translate(translator)

        clean_card_id = re.sub(r'\W+', '', card_id)

        return clean_card_id

    def clean_card(self, card_id: str) -> str:
        """
        Clean the card id with all the necessary cleaning functions

        Parameters
        ----------
        card_id: str
            Input card_id to clean

        Returns
        -------
        str
            The clean card without any special characters or spaces
        """
        clean_card_id = card_id.lower()
        clean_card_id = self.__remove_spaces(clean_card_id)
        clean_card_id = self.__remove_special_characters(clean_card_id)

        return clean_card_id


def extract_null_values(
    data: pd.DataFrame,
    by_col: str
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Extracting NULL values from specific DataFrame

    Parameters
    ----------
    data : pd.DataFrame
        Basic DataFrame
    by_col : str
        Column to separate Null values

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        Tuple of Non-Null DataFrame and Null DataFrame
    """

    null_data = data[data[by_col].isna()]

    non_null_data = data.dropna(subset=[by_col])

    return non_null_data, null_data


def clean_card_data(
    data: pd.DataFrame,
    card_col: str
) -> pd.DataFrame:
    """
    Preprocess card_id to clean format

    Parameters
    ----------
    data : pd.DataFrame
        Basic DataFrame
    card_col : str
        Column contains card_id to clean

    Returns
    -------
    pd.DataFrame
        Original DataFrame with a new columns named clean_<card_col>
    """
    if data.empty:
        return data

    card_cleaner = CardIDCleaner()

    clean_data = data

    clean_data[f'clean_{card_col}'] =\
        clean_data[card_col].apply(card_cleaner.clean_card)

    return clean_data



"""
Module to normalized input string to unicode dung san (NFC)
"""
import unicodedata


def minimal_convert_unicode(txt: str) -> str:
    """
    Convert to NFC - Unicode Composed

    Parameters
    ----------
    txt : str
        The input sentence

    Returns
    -------
    str
        The output sentence converted to NFC format
    """
    normalized_txt = unicodedata.normalize('NFKC', txt)

    return normalized_txt.encode().decode()



import re


def basic_phone_preprocess(phone: str) -> str:
    """
    Perform basic preprocessing on phone

    1. Remove all non-digit from phone string
    2. Remove all the spaces from phone string
    3. Remove (84) from head phone
    4. Add 0 to number not starting with 0

    Parameters
    ----------
    phone : str
        The input phone to preprocess

    Returns
    -------
    str
        Preprocessed phone
    """
    if len(phone) < 9:
        return phone

    clean_phone = re.sub(r"[^0-9]", "", phone)
    clean_phone = re.sub(r"\s+", "", clean_phone)
    if len(clean_phone) < 9:
        return clean_phone

    if len(clean_phone) != 9:
        clean_phone = re.sub(r"(?i)^84", "0", clean_phone)
    if clean_phone[0] != '0':
        clean_phone = '0' + clean_phone

    return clean_phone



import pickle

import tensorflow as tf
from tensorflow.keras import layers


def create_vectorizations(train_pairs, sequence_length=50, vocab_size=15000):
    source_vectorization = layers.TextVectorization(
        max_tokens=vocab_size,
        output_mode='int',
        output_sequence_length=sequence_length
    )
    train_stripped_texts = [n[0] for n in train_pairs]
    source_vectorization.adapt(train_stripped_texts)

    target_vectorization = layers.TextVectorization(
        max_tokens=vocab_size,
        output_mode='int',
        output_sequence_length=sequence_length+1
    )
    train_original_texts = [n[1] for n in train_pairs]
    target_vectorization.adapt(train_original_texts)

    return source_vectorization, target_vectorization


def save_vectorization(vectorization, file_path):
    '''
    Save the config and weights of a vectorization to disk as pickle file,
    so that we can reuse it when making inference.
    '''

    with open(file_path, 'wb') as f:
        f.write(pickle.dumps({'config': vectorization.get_config(),
                              'weights': vectorization.get_weights()}))


def load_vectorization_from_disk(vectorization_path):
    '''
    Load a saved vectorization from disk.
    This method is based on the following answer on Stackoverflow.
    https://stackoverflow.com/a/65225240/4510614
    '''

    with open(vectorization_path, 'rb') as f:
        from_disk = pickle.load(f)
        new_v = layers.TextVectorization(max_tokens=from_disk['config']['max_tokens'],
                                         output_mode='int',
                                         output_sequence_length=from_disk['config']['output_sequence_length'])

        # You have to call `adapt` with some dummy data (BUG in Keras)
        new_v.adapt(tf.data.Dataset.from_tensor_slices(["xyz"]))
        new_v.set_weights(from_disk['weights'])
    return new_v



"""
Module to convert old type phones to their new type
"""
from time import time

import pandas as pd
from tqdm import tqdm

from preprocessing_pgp.phone.const import (
    SUB_PHONE_10NUM,
    SUB_PHONE_11NUM,
    SUB_TELEPHONE_10NUM,
    SUB_TELEPHONE_11NUM,
)
from preprocessing_pgp.phone.utils import basic_phone_preprocess
from preprocessing_pgp.utils import (
    parallelize_dataframe,
    sep_display
)
from preprocessing_pgp.phone.converter import (
    convert_mobi_phone,
    convert_phone_region
)
from preprocessing_pgp.phone.detector import (
    detect_mobi_phone_vendor,
    detect_tele_phone_vendor,
    detect_meaningful_phone
)

# ? ENVIRONMENT SETUP
tqdm.pandas()


# ? CHECK & EXTRACT FOR VALID PHONE
def extract_valid_phone(
    phones: pd.DataFrame,
    phone_col: str = "phone",
    print_info: bool = True
) -> pd.DataFrame:
    """
    Check for valid phone by pattern of head-code and convert the valid-old-code to new-code phone

    Parameters
    ----------
    phones : pd.DataFrame
        The DataFrame contains the phones
    phone_col : str, optional
        The columns which direct to the phones, by default "phone"
    print_info : bool, optional
        Whether to print the information of the run

    Returns
    -------
    pd.DataFrame
        The DataFrame with converted phone column and check if valid or not
    """
    if phones.empty:
        return phones
    # * Split na phone
    na_phones = phones[phones[phone_col].isna()]
    #! Prevent override the origin DF
    f_phones = phones[phones[phone_col].notna()]
    origin_cols = f_phones.columns

    # ? Preprocess phone with basic phone string clean up
    f_phones["phone_clean"] = f_phones[phone_col].apply(
        basic_phone_preprocess)

    if print_info:
        print(
            f"# OF PHONE CLEAN : {f_phones.query(f'phone_clean != {phone_col}').shape[0]}",
            end="\n\n",
        )

        # print("Sample of non-clean phones:")
        # print(f_phones.query(f"phone_clean != {phone_col}"), end="\n\n\n")

    # ? Calculate the phone length for further preprocessing
    f_phones["phone_length"] = f_phones["phone_clean"].map(
        lambda x: len(str(x))
    )

    # ? Phone length validation: currently support phone number with length of 10 and 11.
    # ? Also, phone prefix has to be in the sub-phone dictionary.

    # * Length 10 - New
    mask_valid_new_sub_phone = (f_phones["phone_length"] == 10) & (
        f_phones["phone_clean"].str[:3].isin(SUB_PHONE_10NUM)
    )
    f_phones.loc[
        mask_valid_new_sub_phone,
        ["is_phone_valid", "is_mobi", "is_new_mobi"],
    ] = True

    if print_info:
        print(
            f"# OF MOBI PHONE 10 NUM VALID : {mask_valid_new_sub_phone.sum()}",
            end="\n\n\n",
        )

    # * Length 11 - Old
    mask_valid_old_sub_phone = (f_phones["phone_length"] == 11) & (
        f_phones["phone_clean"].str[:4].isin(SUB_PHONE_11NUM)
    )
    f_phones.loc[
        mask_valid_old_sub_phone,
        ["is_phone_valid", "is_mobi", "is_old_mobi"],
    ] = True

    if print_info:
        print(
            f"# OF MOBI PHONE 11 NUM VALID : {mask_valid_old_sub_phone.sum()}",
            end="\n\n\n",
        )

    # ? Correct phone numbers with old phone number format.
    mask_old_phone_format = f_phones["is_old_mobi"] == True

    f_phones.loc[mask_old_phone_format, "phone_convert"] = f_phones.loc[
        mask_old_phone_format, "phone_clean"
    ].map(convert_mobi_phone)

    if print_info:
        print(
            f"# OF OLD MOBI PHONE CONVERTED : {f_phones['phone_convert'].notna().sum()}")

        # print("Sample of converted MOBI phone:", end="\n\n")
        # print(f_phones.loc[(mask_old_phone_format) &
        #                    (f_phones["phone_convert"].notna())])

    # ? Check for valid tele-phone (old/new)

    # * Length 11 - NEW
    mask_valid_new_tele_phone = (
        (f_phones["phone_length"] == 11)
        & (
            (f_phones["phone_clean"].str[:3].isin(SUB_TELEPHONE_11NUM))
            | (f_phones["phone_clean"].str[:4].isin(SUB_TELEPHONE_11NUM))
        )
        & (f_phones["is_mobi"].isna())
    )
    f_phones.loc[
        mask_valid_new_tele_phone,
        ["is_phone_valid", "is_new_landline"],
    ] = True

    # * Length 10 - OLD
    mask_valid_old_tele_phone = (
        (f_phones["phone_length"] == 10)
        & (
            (f_phones["phone_clean"].str[:3].isin(SUB_TELEPHONE_10NUM))
            | (f_phones["phone_clean"].str[:2].isin(SUB_TELEPHONE_10NUM))
            | (f_phones["phone_clean"].str[:4].isin(SUB_TELEPHONE_10NUM))
        )
        & (f_phones["is_mobi"].isna())
    )
    f_phones.loc[
        mask_valid_old_tele_phone,
        ["is_phone_valid", "is_old_landline"],
    ] = True

    # ? Convert head phone of region from old to new

    mask_old_region_phone = f_phones["is_old_landline"] == True

    if print_info:
        print(f"# OF OLD REGION PHONE : {mask_old_region_phone.sum()}")

    f_phones.loc[mask_old_region_phone, "phone_convert"] = f_phones.loc[
        mask_old_region_phone, "phone_clean"
    ].map(convert_phone_region)

    # if print_info:
    #     print("Sample of converted telephone by region:", end="\n\n")
    #     print(f_phones.loc[(mask_old_region_phone) &
    #                        (f_phones["phone_convert"].notna())])

    # ? Filling NaNs in indicator columns

    fill_cols = [
        "phone_clean",
        "is_phone_valid",
        "is_mobi",
        "is_new_mobi",
        "is_old_mobi",
        "is_new_landline",
        "is_old_landline"
    ]
    f_phones[fill_cols] = f_phones[fill_cols].fillna(False)

    # ? Final preprocessing - Case not changing any head code
    f_phones = f_phones.drop("phone_length", axis=1)
    f_phones.loc[
        f_phones["is_phone_valid"] & f_phones["phone_convert"].isna(
        ), "phone_convert"
    ] = f_phones["phone_clean"]

    if print_info:
        print(
            f"# OF VALID PHONE : {f_phones[f_phones['is_phone_valid']].shape[0]}",
            end="\n\n",
        )
        print(
            f"# OF INVALID PHONE : {f_phones[~f_phones['is_phone_valid']].shape[0]}",
            end="\n\n",
        )

        # print("Sample of invalid phones:", end="\n\n")

    # f_phones.drop(phone_col, axis=1, inplace=True)
    # f_phones.rename(columns={"phone_clean": phone_col}, inplace=True)
    f_phones = f_phones[[*origin_cols, *fill_cols, 'phone_convert']]
    # if print_info:
    #     print(f_phones[~f_phones["is_phone_valid"]].head(10))

    final_phones = pd.concat([f_phones, na_phones])
    final_phones[fill_cols] = final_phones[fill_cols].fillna(False)

    # ? Add Vendor
    valid_mobi_phone_mask = (final_phones['is_phone_valid'] &
                             final_phones['is_mobi'])
    final_phones.loc[
        valid_mobi_phone_mask,
        'phone_vendor'
    ] = final_phones.loc[
        valid_mobi_phone_mask,
        'phone_convert'
    ].apply(detect_mobi_phone_vendor)

    valid_tele_phone_mask = (final_phones['is_phone_valid'] &
                             ~final_phones['is_mobi'])
    final_phones.loc[
        valid_tele_phone_mask,
        'phone_vendor'
    ] = final_phones.loc[
        valid_tele_phone_mask,
        'phone_convert'
    ].apply(detect_tele_phone_vendor)

    # ? Detect meaningful phone
    final_phones.loc[
        final_phones['is_phone_valid'],
        'tail_phone_type'
    ] = final_phones.loc[
        final_phones['is_phone_valid'],
        'phone_convert'
    ].apply(detect_meaningful_phone)

    # ? Add phone type
    final_phones.loc[
        final_phones['is_mobi'],
        'phone_type'
    ] = 'mobile phone'
    final_phones.loc[
        (~final_phones['is_mobi']) & (final_phones['is_phone_valid']),
        'phone_type'
    ] = 'landline'

    return final_phones


def process_convert_phone(
    data: pd.DataFrame,
    phone_col: str = 'phone',
    n_cores: int = 1,
    logging_info: bool = True
) -> pd.DataFrame:
    """
    Converting valid phone to new phone type

    Parameters
    ----------
    data : pd.DataFrame
        The data containing the phone numbers
    phone_col : str, optional
        The column name contains records of phones, by default 'phone'
    n_cores : int, optional
        The number of core to process, by default 1

    Returns
    -------
    pd.DataFrame
        The converted data with new columns:
        * `is_phone_valid`: indicator for valid phone
        * `phone_type`: the phone type (mobi or landline)
        * `phone_convert`: converted valid old phone type to new phone type
        * `phone_vendor`: the vendor type of the phone number
        * `tail_phone_type`: the tail phone number meanings
    """

    # * Select only phone column
    orig_cols = data.columns
    phone_data = data[[phone_col]]

    # * Validate and convert phone
    if logging_info:
        print(">>> Converting phones: ")
    start_time = time()

    converted_data = parallelize_dataframe(
        phone_data,
        extract_valid_phone,
        n_cores=n_cores,
        phone_col=phone_col,
        print_info=False
    )

    convert_time = time()-start_time

    if logging_info:
        print(f"{int(convert_time)//60}m{int(convert_time)%60}s")

    # * Concat with original cols
    new_cols = [
        'phone_clean',
        'is_phone_valid',
        'phone_type',
        'phone_convert',
        'phone_vendor',
        'tail_phone_type'
    ]
    converted_data = converted_data[new_cols]

    converted_data = pd.concat([
        data[orig_cols],
        converted_data
    ], axis=1)

    return converted_data



"""
Module contains validator for email
"""

import re
from time import time

import pandas as pd

from preprocessing_pgp.email.utils import (
    split_email,
    is_name_accented,
    is_valid_email_domain
)
from preprocessing_pgp.email.const import (
    AT_LEAST_ONE_CHAR_REGEX,
    LEAST_NUM_EMAIL_CHAR,
    EMAIL_DOMAIN_REGEX,
    COMMON_EMAIL_REGEX,
    EDGE_AUTO_EMAIL_REGEX,
    PRIVATE_EMAIL_DOMAINS,
    DOMAIN_GROUP_DICT,
    EDU_EMAIL_REGEX
)
from preprocessing_pgp.email.preprocess import (
    clean_email
)
from preprocessing_pgp.utils import (
    parallelize_dataframe
)

pd.options.mode.chained_assignment = None


class EmailValidator:
    """
    Class contains building components for validating email
    """

    def __init__(
        self,
        domain_dict: pd.DataFrame = None
    ) -> None:
        self.email_services = EMAIL_DOMAIN_REGEX.values()
        self.domain_dict = domain_dict

    def is_valid_email(
        self,
        email: str = None
    ) -> bool:
        """
        Check whether the email is valid or not

        Parameters
        ----------
        email : str
            The input email to check for validation

        Returns
        -------
        bool
            Whether the input email is valid with the basic email rules
        """

        if not email:
            return False

        normed_email = email.lower()
        email_name, email_group = split_email(email)

        # if not is_valid_email_domain(email_group):
        #     return False

        if self.is_auto_email(normed_email):
            return False

        large_company_domain = self.get_large_company_domain(email_group)
        if large_company_domain is not None:
            return self._is_valid_email_regex(
                email_name,
                EMAIL_DOMAIN_REGEX[large_company_domain]['regex']
            )

        if self.is_student_email(email_group):
            return self._is_valid_email_regex(
                email_name,
                EDU_EMAIL_REGEX
            )

        return self.is_common_email(normed_email)

    def _is_valid_email_regex(
        self,
        email: str,
        email_regex: str
    ) -> bool:
        """
        Check whether the email match the regex
        """
        return bool(re.match(email_regex, email))

    def get_large_company_domain(
        self,
        email_group: str
    ) -> str:
        """
        Check whether the email is of large company email:
        1. gmail
        2. yahoo
        3. microsoft
        4. fpt

        Parameters
        ----------
        email_group : str
            The input email_group to get the domain name

        Returns
        -------
        str
            The email group that email was in
        """
        for domain_regex, domain_name in DOMAIN_GROUP_DICT.items():
            if re.match(domain_regex, email_group):
                return domain_name

        return None

    def is_common_email(
        self,
        email: str
    ) -> bool:
        """
        Check whether the email is common email account or not
        """

        if re.match(COMMON_EMAIL_REGEX, email):
            # email_name, _ = split_email(email)
            # return self._is_valid_email_name(email_name)
            return True
        return False

    def is_student_email(
        self,
        email_group: str
    ) -> bool:
        """
        Check whether the email is of student email or not
        """
        if not email_group:
            return False

        return bool(re.search('edu', email_group))

    def is_auto_email(
        self,
        email: str
    ) -> bool:
        """
        Check whether the email is auto-email or not
        """
        return bool(re.match(EDGE_AUTO_EMAIL_REGEX, email))

    def _is_valid_email_name(
        self,
        email_name: str
    ) -> bool:
        """
        Check valid email's name with 3 criteria:
        1. Non-accented
        2. At least 8 characters
        3. Contains at least 1 alpha character

        Parameters
        ----------
        email_name : str
            The input email's name

        Returns
        -------
        bool
            Whether the email name is valid or not
        """

        return not is_name_accented(email_name)\
            and self.__is_valid_name_length(email_name)\
            and self.__is_valid_name_syntax(email_name)

    def __is_valid_name_length(
        self,
        email_name: str
    ) -> bool:
        """
        Check whether the email's name length is in valid range
        """
        return len(email_name) >= LEAST_NUM_EMAIL_CHAR

    def __is_valid_name_syntax(
        self,
        email_name: str
    ) -> bool:
        """
        Check whether the email's name is at valid syntax
        """
        return bool(re.match(AT_LEAST_ONE_CHAR_REGEX, email_name))

    def check_valid_email_domain(
        self,
        data: pd.DataFrame,
        domain_col: str = 'email_domain'
    ) -> pd.DataFrame:
        """
        Checking whether the email domain is valid or not
        """
        if self.domain_dict is None:
            data['is_domain_valid'] =\
                data[domain_col].apply(is_valid_email_domain)
            return data

        if not self.domain_dict.columns.isin(['email_domain', 'is_domain_valid']).all():
            print("Domain dictionary not match schema, re-running all domains")
            data['is_domain_valid'] =\
                data[domain_col].apply(is_valid_email_domain)
            return data

        new_domains = list(set(data[domain_col].unique())-set(self.domain_dict['email_domain']))
        new_domain_df = pd.DataFrame({
            'email_domain': new_domains
        })
        new_domain_df['is_domain_valid'] \
            = new_domain_df['email_domain'].apply(is_valid_email_domain)

        domain_mapper = pd.concat([self.domain_dict, new_domain_df], ignore_index=True)
        data = pd.merge(
            data.set_index('email_domain'),
            domain_mapper.set_index('email_domain'),
            left_index=True, right_index=True,
            how='left', sort=False
        ).reset_index()

        return data


def validate_clean_email(
    data: pd.DataFrame,
    email_col: str = 'cleaned_email',
    domain_dict: pd.DataFrame = None
) -> pd.DataFrame:
    """
    Process validating for cleaned email
    and output an additional column for validator: `is_email_valid`

    Parameters
    ----------
    data : pd.DataFrame
        The dataframe contains cleaned email column
    email_col : str, optional
        The cleaned email column name, by default 'cleaned_email'
    domain_dict : pd.DataFrame, optional
        The domain dictionary to use for faster reference, by default None

    Returns
    -------
    pd.DataFrame
        Validated data contains the email's valid indicator: `is_email_valid`
    """
    if data.empty:
        return data

    validator = EmailValidator(domain_dict=domain_dict)

    validated_data = data

    # * Check for email domain
    validated_data['email_domain'] =\
        validated_data[email_col]\
        .str.split('@').str[1]
    validated_data = validator.check_valid_email_domain(
        validated_data,
        domain_col='email_domain'
    )
    valid_domain_data = validated_data.query('is_domain_valid')
    invalid_domain_data = validated_data.query('~is_domain_valid')

    # * Check for email regex
    valid_domain_data['is_email_valid'] = valid_domain_data[email_col].apply(
        validator.is_valid_email
    )

    # * Concat with invalid domain data
    invalid_domain_data['is_email_valid'] = False
    validated_data = pd.concat([valid_domain_data, invalid_domain_data])

    # * Check for autoemail
    validated_data['is_autoemail'] = validated_data[email_col].apply(
        validator.is_auto_email
    )

    return validated_data


def process_validate_email(
    data: pd.DataFrame,
    email_col: str = 'email',
    n_cores: int = 1,
    domain_dict: pd.DataFrame = None
) -> pd.DataFrame:
    """
    Process validating email address

    Parameters
    ----------
    data : pd.DataFrame
        The dataframe contains email records
    email_col : str, optional
        The column name that hold email records, by default 'email'
    n_cores : int, optional
        The number of cores used to run parallel, by default 1 core will be used
    domain_dict : pd.DataFrame, optional
        The domain dictionary to use for faster reference, by default None

    Returns
    -------
    pd.DataFrame
        The data with additional columns:
        * `is_email_valid`: indicator for whether the email is valid or not
        * `is_autoemail`: indicator for whether the email is autoemail or not
    """
    # * Select only email column to continue
    orig_cols = data.columns
    email_data = data[[email_col]]

    # * Cleansing email
    print(">>> Cleansing email: ", end='')
    start_time = time()
    cleaned_data = parallelize_dataframe(
        email_data,
        clean_email,
        n_cores=n_cores,
        email_col=email_col
    )
    clean_time = time() - start_time
    print(f"{int(clean_time)//60}m{int(clean_time)%60}s")

    # * Separate na data
    na_data = cleaned_data[cleaned_data[f'cleaned_{email_col}'].isna()]
    non_na_data = cleaned_data[cleaned_data[f'cleaned_{email_col}'].notna()]

    # * Validating email
    print(">>> Validating email: ", end='')
    start_time = time()
    validated_data = parallelize_dataframe(
        non_na_data,
        validate_clean_email,
        n_cores=n_cores,
        email_col=f'cleaned_{email_col}',
        domain_dict=domain_dict
    )
    validate_time = time() - start_time
    print(f"{int(validate_time)//60}m{int(validate_time)%60}s")

    # * Get the domain of the email name & Check for private email
    print(">>> Get email domain: ", end='')
    start_time = time()
    validated_data['email_domain'] = validated_data['email_domain'].replace(
        DOMAIN_GROUP_DICT, regex=True)
    validated_data['private_email'] = validated_data['email_domain'].isin(
        PRIVATE_EMAIL_DOMAINS)
    domain_time = time() - start_time
    print(f"{int(domain_time)//60}m{int(domain_time)%60}s")

    # * Concat with the nan data
    final_data = pd.concat([validated_data, na_data])

    # * Filling na data to invalid email
    final_data['is_email_valid'].fillna(False, inplace=True)
    final_data['is_email_valid'] = final_data['is_email_valid'].astype(bool)

    # * Concat with the origin cols
    new_cols = [
        f'cleaned_{email_col}',
        'is_email_valid',
        'is_autoemail',
        'email_domain',
        'is_domain_valid',
        'private_email'
    ]
    final_data = pd.concat([data[orig_cols], final_data[new_cols]], axis=1)

    return final_data



"""
Module for essential constants in enrich name
"""
import os

import pandas as pd

# ? MODEL PATHS
NAME_SPLIT_PATH = os.path.join(
    os.path.dirname(__file__),
    os.pardir,
    'data',
    'name_split'
)

MODEL_PATH = os.path.join(
    os.path.dirname(__file__),
    os.pardir,
    'data',
    'transformer_model',
    'trial-25'
)

RULE_BASED_PATH = os.path.join(
    os.path.dirname(__file__),
    os.pardir,
    'data',
    'rule_base',
    'students'
)

NICKNAME_PATH = os.path.join(
    os.path.dirname(__file__),
    '../data/name_split/nicknames_boundary.parquet'
)

GENDER_MODEL_PATH = os.path.join(
    os.path.dirname(__file__),
    '../data/gender_model'
)

# ? PREPROCESS CONSTANTS
NON_HUMAN_REG_LIST = [
    # Companies
    'Cong Ty', 'Co Phan', 'Co Phieu', 'Bds', 'Bat Dong San', 'Tnhh', 'Thuong Nghiep', 'Huu Hang',
    # Schools
    # 'Doan Truong', 'Van Phong',
    'Truong Hoc', 'Thpt', 'Thcs', 'Dai Hoc', 'Tieu Hoc', 'Noi Tru', 'Ngoai Tru', 'Cao Hoc', 'Thac Si', 'Tien Si', 'Phd', 'Pho Giao Su', 'Hoc Vien',
    # Have Meanings
    'To Phuong', 'Nghi Viec', 'Thoi Viec', 'Lam Viec', 'So Y Te', 'Ve Sinh',
    # Others
    'Ca Bong', 'Ca Chep'
]

'''
NON-OFFICIAL NAMES : Ref from 'JOURNAL OF ETHNIC MINORITIES RESEARCH - 2019'

- NON_OFFICIAL_NAMES: Tên không chính thức có dấu của các dân tộc ở Việt Nam.
- LOCAL_GROUP_NAMES: Tên chỉ các nhóm địa phương, nhóm riêng.
'''
NON_OFFICIAL_NAMES = [
    # Kinh
    'Keo', 'Doan',
    # Khmer
    'Cur', 'Cul', 'Khơ Me', 'Krôm',
    # Mường
    'Mol', 'Mual', 'Moi'
    # Mông
    'Mèo', 'Mẹo'
    # Dao
    'Mán', 'Động', 'Trại', 'Dìu Miền', 'Kiềm Miền',
    # Ngái
    'Xín', 'Lê', 'Đán', 'Khách Gia',
    # Sán Chay
    'Hờn Bạn', 'Sơn Tử',
    # Cơ Ho
    'Còn Chau',
    # Chăm
    'Chàm', 'Chiêm',
    # Sán Dìu
    'Trại', 'Trại Đất', 'Mán Quần Cộc',
    # Hrê
    'Chăm Hrê', 'Chom, Lũy',
    # Mnông
    'Pnông', 'Nông',
    # Giáy
    'Nhắng', 'Dẩng',
    # Gié - Triêng
    'Giang Rẫy',
    # Mạ
    'Còn Chau', 'Chau Mạ',
    # Khơ Mú
    'Xá', 'Xá Cẩu', 'Mứn Xen', 'Pu Thênh', 'Tềnh', 'Tày Hạy'
]

LOCAL_GROUP_NAMES = [
    # Tày
    'Tày', 'Ngạn', 'Pa Dí', 'Thu Lao',
    # Gia Rai
    'Chor', 'Hđrung', 'Arap', 'Mthur', 'Tbuăn',
    # Chứt
    'Arem', 'Rục', 'Mày', 'Sách', 'Mã Liềng', 'Kri',
    # Ê Đê
    'Kpă', 'Mthur', 'Ktul', 'Đliê', 'Hruê', 'Blô', 'Ê Pan', 'Bih', 'Krung', 'Kđrao',
    # Thổ
    'Thổ', 'Kẹo', 'Họ', 'Mọn', 'Cuối', 'Tày Poọng', 'Đan Lai',
    # Xơ Đăng
    'Xơ Teng', 'Hđang', 'Tơ Đrá', 'Mơ Nâm', 'Ha Lăng', 'Ca Dong',
    # Bru - Vân Kiều
    'Khùa', 'Mang Coong', 'Trì', 'Sộ', 'Vân Kiều',
    # Giẻ - Triêng
    'Giẻ', 'Triêng', 'Ta Liêng', 'Ve', 'Bơ Noong', 'Pơ Noong'
]


REPLACE_HUMAN_REG_DICT = {
    'K HI': '',
    'Bs': '',
    'Ng.': 'Nguyễn'
}

BRIEF_NAME_DICT = {
    'nguyen': ['ng.', 'n.'],
    'do': ['d.'],
    'pham': ['p.']
}

# * NICKNAMES
NICKNAMES = pd.read_parquet(NICKNAME_PATH)
NICKNAME_REGEX = '|'.join(
    [*NICKNAMES['name'].to_list(),
     *NICKNAMES[NICKNAMES['de_name'].str.split().str.len() > 1]['de_name'].to_list()]
)



import re
from time import time

import numpy as np
import pandas as pd
import multiprocessing as mp
from unidecode import unidecode
from flashtext import KeywordProcessor
from tqdm import tqdm

from preprocessing_pgp.name.const import (
    BRIEF_NAME_DICT,
    NAME_SPLIT_PATH
)

tqdm.pandas()


def BuildLastName(base_path):
    # load stats lastname
    stats_lastname_vn = pd.read_parquet(
        f'{base_path}/stats_lastname_vn.parquet')

    last_name_list1 = list(
        stats_lastname_vn[stats_lastname_vn['No'] <= 50]['Last_Name'].unique())
    last_name_list2 = ['Nguyễn', 'Trần', 'Lê', 'Phạm', 'Hoàng', 'Huỳnh', 'Phan',
                       'Vũ', 'Võ', 'Đặng', 'Bùi', 'Đỗ', 'Hồ', 'Ngô', 'Dương', 'Lý',
                       'Trương', 'Bùi', 'Đinh', 'Lương', 'Tạ', 'Quách', 'Hứa']
    last_name_list3 = last_name_list1
    last_name_list = list(set(last_name_list3) | set(
        last_name_list1) | set(last_name_list2))

    # return
    return last_name_list1, last_name_list2, last_name_list3, last_name_list


def BuildWordName(base_path):
    # load database name
    ext_data_name1 = pd.read_parquet(f'{base_path}/ext_data.parquet')
    ext_data_name2 = pd.read_parquet(f'{base_path}/ext_data_uit.parquet')

    ext_data_name1.columns = ['full_name', 'gender']
    ext_data_name1['full_name'] = ext_data_name1['full_name'].str.replace(
        '\s+', ' ', regex=True).str.title()
    ext_data_name1['first_name'] = ext_data_name1['full_name'].str.split(
        ' ').str[-1]
    ext_data_name1['last_name_group'] = ext_data_name1['full_name'].str.split(
        ' ').str[0]
    ext_data_name1['last_name'] = ext_data_name1['full_name'].str.split(
        ' ').str[:-1].str.join(' ')

    ext_data_name2['full_name'] = ext_data_name2['full_name'].str.replace(
        '\s+', ' ', regex=True).str.title()
    ext_data_name2 = ext_data_name2[[
        'full_name', 'gender', 'first_name', 'last_name_group', 'last_name']].copy()

    # stats freq
    ext_data_name1['full_name_unicecode'] = ext_data_name1['full_name'].str.lower(
    ).apply(unidecode)
    stats_word_name1 = ext_data_name1['full_name_unicecode'].str.split(
        expand=True).stack().value_counts().reset_index()
    stats_word_name1.columns = ['Word', 'Frequency']

    ext_data_name2['full_name_unicecode'] = ext_data_name2['full_name'].str.lower(
    ).apply(unidecode)
    stats_word_name2 = ext_data_name2['full_name_unicecode'].str.split(
        expand=True).stack().value_counts().reset_index()
    stats_word_name2.columns = ['Word', 'Frequency']

    # wordName
    stats_word_name = pd.concat(
        [stats_word_name1, stats_word_name2], axis=0, ignore_index=False)
    # stats_word_name = stats_word_name1.append(
    # stats_word_name2, ignore_index=False)
    stats_word_name = stats_word_name.groupby(
        by=['Word'])['Frequency'].sum().reset_index()
    stats_word_name = stats_word_name[~((stats_word_name['Frequency'] < 5) |
                                        stats_word_name['Word'].str.contains('[^a-z]'))]
    word_name = set(stats_word_name['Word'])

    # return
    return word_name


class NameProcess:
    def __init__(self):

        self.word_name = np.array(list(BuildWordName(NAME_SPLIT_PATH)))
        self.last_name_list1, self.last_name_list2, self.last_name_list3, self.last_name_list = BuildLastName(
            NAME_SPLIT_PATH)
        self.__generate_brief_keywords()
        self.brief_name_terms = list(
            self.brief_name_kws.get_all_keywords().keys())

    def __generate_brief_keywords(self):
        if hasattr(self, 'brief_name_kws'):
            return

        kws = KeywordProcessor(case_sensitive=True)
        kws.add_keywords_from_dict(BRIEF_NAME_DICT)

        self.brief_name_kws = kws

    def CountNameVN(self, text):
        try:
            text = unidecode(text.lower())

            # Contains 1 char
            if len(text) == 1:
                return 1

            # Check in word_name
            word_text = np.array(text.split())

            # Handle case where 2 word with same decode in name
            intersect_bool = np.in1d(word_text, self.word_name)
            intersect_score = np.sum(intersect_bool)

            # name containing brief 1 word and brief name
            brief_score = np.sum(np.in1d(word_text, self.brief_name_terms))

            return intersect_score + brief_score

        except:
            return -1

    def check_name_valid(
        self,
        series: pd.Series
    ) -> pd.Series:
        de_data = series\
            .apply(unidecode)\
            .str.lower()\
            .str.split(expand=True)

        name_valid = de_data.apply(
            lambda row:
                row.isin(self.word_name).all()
        )

        return name_valid

    def CleanName(self, raw_name):
        try:
            process_name = raw_name.lower().strip()

            # is email?
            regex_email = r'[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+'
            process_name = re.sub(regex_email, '', process_name)
            # print(f'email char: {process_name}')

            # is phone?
            regex_phone = r'[0-9]*'
            process_name = re.sub(regex_phone, '', process_name)

            # special char
            process_name = re.sub(r'\(.*\)', '', process_name)
            process_name = re.sub(r'\.|,|-|_', '', process_name)
            process_name = process_name.split(',')[0]
            process_name = process_name.split('-')[0]
            process_name = process_name.strip('-| |.|,|(|)')
            # print(f'special char: {process_name}')

            # fix typing
            regex_typing1 = '0[n|a|i|m|c|t][^wbmc\d ]'
            regex_typing2 = '[h|a|b|v|g|e|x|c]0[^0-9.-]'
            process_name = re.sub(regex_typing1, 'o', process_name)
            process_name = re.sub(regex_typing2, 'o', process_name)
            # print(f'fix typing char: {process_name}')

            # pronoun
            regex_pronoun1 = r'^(?:\bkh\b|\bkhach hang\b|\bchị\b|\bchi\b|\banh\b|\ba\b|\bchij\b|\bc\b|\be\b|\bem\b|\bcô\b|\bco\b|\bchú\b|\bbác\b|\bbac\b|\bme\b|\bdì\b|\bông\b|\bong\b|\bbà\b|\ba\.|\bc\.)\s+'
            regex_pronoun2 = r'^(?:\bnội\b|\bngoại\b)\s+'
            regex_pronoun3 = r'^(?:\bvo anh\b|\bvo a\b|\bvo chu\b|\bbo anh\b|\bme anh\b|\bem anh\b|\bbo a\b|\bban chi\b|\bbo chi\b|\bban\b|\bck\b|\bvk\b)\s+'
            try:
                pronouns = re.findall(regex_pronoun1, process_name)
                pronoun = pronouns[0].strip()
            except:
                pronoun = None
            process_name = re.sub(regex_pronoun1, '', process_name)
            process_name = re.sub(regex_pronoun2, '', process_name)
            process_name = re.sub(regex_pronoun3, '', process_name)
            process_name = process_name.strip('-| |.|,|(|)')

            # defaut
            regex_default1 = r'người mua|người nhận|số ngoài danh bạ|nhập số điện thoại|giao hàng|test|[~!#$%^?]'
            regex_default2 = r'dung chung|ky thuat|du phong|dong tien|chu cu|chung phong|co quan|thu nhat|thu hai|so cong ty|chu moi|nhan su|dong nghiep|lien quan|em cua|may ban|so may|nghi lam|quan ly|dat so|su dung|nhan vien|chu nha|moi mua|dien thoai|chuyen di|lap dat|cung phong|nham so|hop dong|tong dai|can ho|ke toan|k co|so phu|lien he|lien lac|don di|so cu|so moi|ve que|khong dung|ben canh|ko co'
            regex_default3 = r'kh |khach hang|chu nha|cong ty|cc|nsd|vo kh|chong kh|so chu hd|so moi|nguoi moi tiep nhan'
            process_name = re.sub(regex_default1, '', process_name)
            process_name = re.sub(regex_default2, '', process_name)
            process_name = re.sub(regex_default3, '', process_name)

            process_name = process_name.strip('-| |.|,|(|)')
            process_name = re.sub(r'\s+', ' ', process_name)
            # print(f'remove noun char: {process_name}')

            # is dict name VN
            pct_vn = self.CountNameVN(process_name) / \
                len(process_name.split(' '))
#             print(pct_vn)
            process_name = None if ((pct_vn < 0.5) |
                                    (len(process_name) == 1) |
                                    (len(process_name.split(' ')) > 6)
                                    ) else process_name
            # print(f'check VN char: {process_name}')

            # * Replace brief terms
            if process_name is not None:
                process_name = self.brief_name_kws.replace_keywords(
                    process_name)

            # title
            process_name = process_name.title()

            return process_name, pronoun
        except:
            return None, None

    # SPLIT_NAME

    def SplitName(self, full_name):
        try:
            # Variable
            use_only_last_name = True
            full_name = full_name.replace('\s+', ' ').strip().title()
            last_name = ''
            middle_name = None
            first_name = None

            # Case 0: B
            if (full_name.split(' ') == 1):
                # print('Case 0')
                first_name = full_name
                return last_name, middle_name, first_name

            # Case 1: Nguyen Van C
            check_end_case1 = False
            while (check_end_case1 == False):
                for key_vi in self.last_name_list1:
                    key_vi = key_vi + ' '

                    is_case11 = (full_name.find(key_vi) == 0)
                    is_case12 = (full_name.find(unidecode(key_vi)) == 0)
                    is_case1 = is_case11 or is_case12
                    if is_case1:
                        # print('Case 1')
                        key = key_vi if is_case11 else unidecode(key_vi)

                        last_name = (last_name + ' ' + key).strip()
                        full_name = full_name.replace(key, '', 1).strip()

                        if (use_only_last_name == True):
                            check_end_case1 = True
                        break

                    if (full_name.split(' ') == 1) or (key_vi.strip() == self.last_name_list1[-1]):
                        check_end_case1 = True

            # Case 2: Van D Nguyen
            if (last_name.strip() == ''):
                check_end_case2 = False
                while (check_end_case2 == False):
                    for key_vi in self.last_name_list2:
                        key_vi = ' ' + key_vi

                        is_case21 = (len(full_name)-full_name.rfind(key_vi)
                                     == len(key_vi)) & (full_name.rfind(key_vi) != -1)
                        is_case22 = (len(full_name)-full_name.rfind(unidecode(key_vi)) == len(
                            unidecode(key_vi))) & (full_name.rfind(unidecode(key_vi)) != -1)

                        is_case2 = is_case21 or is_case22
                        if is_case2:
                            # print('Case 2')
                            key = key_vi if is_case21 else unidecode(key_vi)

                            last_name = (key + ' ' + last_name).strip()
                            full_name = ''.join(
                                full_name.rsplit(key, 1)).strip()

                            if (use_only_last_name == True):
                                check_end_case2 = True
                            break

                        if (full_name.split(' ') == 1) or (key_vi.strip() == self.last_name_list2[-1]):
                            check_end_case2 = True

            # Case 3: E Nguyen Van
            if (last_name.strip() == ''):
                temp_full_name = full_name
                temp_first_name = temp_full_name.split(' ')[0]
                temp_full_name = ' '.join(
                    temp_full_name.split(' ')[1:]).strip()

                check_end_case3 = False
                while (check_end_case3 == False):
                    for key_vi in self.last_name_list3:
                        key_vi = key_vi + ' '

                        is_case31 = (temp_full_name.find(key_vi) == 0)
                        is_case32 = (temp_full_name.find(
                            unidecode(key_vi)) == 0)
                        is_case3 = is_case31 or is_case32
                        if is_case3:
                            # print('Case 3')
                            key = key_vi if is_case31 else unidecode(key_vi)

                            last_name = (last_name + ' ' + key).strip()
                            temp_full_name = temp_full_name.replace(
                                key, '', 1).strip()

                            if (use_only_last_name == True):
                                check_end_case3 = True
                            break

                        if (full_name.split(' ') == 1) or (key_vi.strip() == self.last_name_list3[-1]):
                            check_end_case3 = True

                if (last_name.strip() != ''):
                    first_name = temp_first_name
                    middle_name = temp_full_name

                    return last_name, middle_name, first_name

            # Fillna
            first_name = full_name.split(' ')[-1]
            try:
                full_name = ''.join(full_name.rsplit(first_name, 1)).strip()
                middle_name = full_name
            except:
                # print('Case no middle name')
                middle_name = None

            # Replace '' to None
            last_name = None if (last_name == '') else last_name
            middle_name = None if (middle_name == '') else middle_name
            first_name = None if (first_name == '') else first_name

            return last_name, middle_name, first_name

        except:
            return None, None, None

    def unidecode_with_na(self, name):
        if name is None:
            return 'NONE'
        return unidecode(name)

    def CoreBestName(self, raw_names_n, name_col='name', key_col='phone'):
        start_time = time()
        raw_names_n = raw_names_n[raw_names_n[name_col].notna()]
        # Rename name column
        raw_names_n = raw_names_n.rename(columns={name_col: 'raw_name'})
        # Skip name (non personal)
        map_name_customer = raw_names_n[['raw_name']].copy().drop_duplicates()
        map_name_customer.columns = ['name']
        map_name_customer['num_word'] = map_name_customer['name'].str.split(
            ' ').str.len()
        skip_names = map_name_customer[(
            map_name_customer['num_word'] > 5)]['name'].unique()
        skip_names_df = raw_names_n[raw_names_n['raw_name'].isin(
            skip_names)][[key_col, 'raw_name']].copy().drop_duplicates()
        names_df = raw_names_n[~raw_names_n['raw_name'].isin(
            skip_names)][[key_col, 'raw_name']].copy().drop_duplicates()
        print(">> Skip/Filter name")
        # Split name: last, middle, first
        map_split_name = names_df[['raw_name']].copy().drop_duplicates()
        with mp.Pool(8) as pool:
            map_split_name[['last_name', 'middle_name', 'first_name']
                           ] = pool.map(self.SplitName, map_split_name['raw_name'])
        names_df = names_df.merge(map_split_name, how='left', on=['raw_name'])
        # Create group_id -> by firstname
        print(names_df[names_df['first_name'].isna()])
        names_df['unidecode_first_name'] = names_df['first_name'].progress_apply(
            self.unidecode_with_na)
        names_df['group_id'] = names_df[key_col] + \
            '-' + names_df['unidecode_first_name']
        names_df = names_df.drop(columns=['unidecode_first_name'])
        # Split case process best_name
        names_df.loc[names_df['last_name'].notna(),
                     'unidecode_last_name'] = names_df.loc[names_df['last_name'].notna(), 'last_name'].apply(unidecode)
        names_df['num_last_name'] = names_df.groupby(
            by=['group_id'])['unidecode_last_name'].transform('nunique')
        names_df['mode_last_name'] = names_df.groupby(
            by=['group_id'])['unidecode_last_name'].progress_transform(lambda x: x.mode())
#         names_df['type'] = names_df.groupby(by=['group_id'])['mode_last_name'].transform(lambda x: x.apply(lambda y:  if y==None))
        info_name_columns = ['group_id', 'raw_name',
                             'last_name', 'middle_name', 'first_name']
        # In TESTING PROGRESS
        n_lastname_mask = (names_df['num_last_name'] >= 2) \
            & (names_df['mode_last_name'] != names_df['unidecode_last_name']) \
            & (names_df['unidecode_last_name'] != None)
        names_n_df = names_df[n_lastname_mask][info_name_columns].copy()
        names_1_df = names_df[~n_lastname_mask][info_name_columns].copy()
#         return names_n_df, names_1_df
        print(">> Create group_id")
        # Process case: 1 first_name - n last_name
        post_names_n_df = names_n_df[names_n_df['last_name'].isna()].copy()
        map_names_n_df = names_n_df[names_n_df['last_name'].notna() &
                                    names_n_df['group_id'].isin(post_names_n_df['group_id'])].copy()
        map_names_n_df['num_char'] = map_names_n_df['raw_name'].str.len()
        map_names_n_df['num_word'] = map_names_n_df['raw_name'].str.split(
            ' ').str.len()
        map_names_n_df['accented'] = map_names_n_df['raw_name'] != map_names_n_df['raw_name'].apply(
            unidecode)
        map_names_n_df = map_names_n_df.sort_values(
            by=['group_id', 'num_word', 'num_char', 'accented'], ascending=False)
        map_names_n_df = map_names_n_df.groupby(by=['group_id']).head(1)
        map_names_n_df = map_names_n_df[['group_id', 'raw_name']].rename(
            columns={'raw_name': 'best_name'})
        post_names_n_df = post_names_n_df.merge(
            map_names_n_df, how='left', on=['group_id'])
        post_names_n_df = post_names_n_df[[
            'group_id', 'raw_name', 'best_name']]
        names_n_df = names_n_df.merge(post_names_n_df, how='left', on=[
                                      'group_id', 'raw_name'])
        names_n_df.loc[names_n_df['best_name'].isna(
        ), 'best_name'] = names_n_df['raw_name']
        names_n_df = names_n_df[['group_id', 'raw_name', 'best_name']]
        print(">> 1 first_name - n last_name")
        # Process case: 1 first_name - 1 last_name
        map_names_1_df = names_1_df[['group_id']].drop_duplicates()
        for element_name in tqdm(['last_name', 'middle_name', 'first_name']):
            # filter data detail
            map_element_name = names_1_df[names_1_df[element_name].notna(
            )][['group_id', element_name]].copy().drop_duplicates()
            # create features
            map_element_name[f'unidecode_{element_name}'] = map_element_name[element_name].apply(
                unidecode)
            map_element_name['num_overall'] = map_element_name.groupby(
                by=['group_id', f'unidecode_{element_name}'])[element_name].transform('count')
            map_element_name = map_element_name.drop(
                columns=f'unidecode_{element_name}')
            map_element_name['num_char'] = map_element_name[element_name].str.len()
            map_element_name['num_word'] = map_element_name[element_name].str.split(
                ' ').str.len()
            map_element_name['accented'] = map_element_name[element_name] != map_element_name[element_name].apply(
                unidecode)
            # approach to choice best
            # map_element_name = map_element_name.sort_values(by=['group_id', 'num_overall', 'num_char', 'num_word', 'accented'], ascending=False)
#             map_element_name = map_element_name.sort_values(
#                 by=['group_id', 'num_word', 'accented', 'num_char', 'num_overall'], ascending=False)
            if element_name != 'middle_name':
                map_element_name = map_element_name.sort_values(
                    by=['group_id',
                        'accented',
                        'num_char',
                        'num_word',
                        'num_overall'], ascending=False)
            else:  # For middlename we choose the longest name then accent latter
                map_element_name = map_element_name.sort_values(
                    by=['group_id',
                        'num_word',
                        'accented',
                        'num_char',
                        'num_overall'], ascending=False)
            map_element_name = map_element_name.groupby(
                by=['group_id']).head(1)
            map_element_name = map_element_name[['group_id', element_name]]
            map_element_name.columns = ['group_id', f'best_{element_name}']
            # merge
            map_names_1_df = map_names_1_df.merge(
                map_element_name, how='left', on=['group_id'])
            map_names_1_df.loc[map_names_1_df[f'best_{element_name}'].isna(
            ), f'best_{element_name}'] = None
        # combine element name
#         return map_names_1_df
        dict_trash = {'': None, 'Nan': None, 'nan': None, 'None': None,
                      'none': None, 'Null': None, 'null': None, "''": None}
        columns = ['best_last_name', 'best_middle_name', 'best_first_name']
        # BIG BUG WHEN NO ENTRY
        map_names_1_df['best_name'] = map_names_1_df[columns].fillna('').agg(' '.join, axis=1).str.replace(
            '(?<![a-zA-Z0-9]),', '', regex=True).str.replace('-(?![a-zA-Z0-9])', '', regex=True)
        # END BUG
        map_names_1_df['best_name'] = map_names_1_df['best_name'].str.strip().replace(
            dict_trash)
        map_names_1_df.loc[map_names_1_df['best_name'].isna(),
                           'best_name'] = None
        # merge
        names_1_df = names_1_df.merge(
            map_names_1_df[['group_id', 'best_name']], how='left', on=['group_id'])
        names_1_df = names_1_df[['group_id', 'raw_name', 'best_name']]
        print(">> 1 first_name - 1 last_name")
        # Concat
        names_df = pd.concat([names_1_df, names_n_df], ignore_index=True)
        # Calculate similarity_score
        name_list_1 = list(names_df['raw_name'].unique())
        name_list_2 = list(names_df['best_name'].unique())
        map_element_name = pd.DataFrame()
        map_element_name['name'] = list(set(name_list_1) | set(name_list_2))
        with mp.Pool(8) as pool:
            map_element_name[['last_name', 'middle_name', 'first_name']] = pool.map(
                self.SplitName, map_element_name['name'])
        for flag in ['raw', 'best']:
            temp = map_element_name.copy()
            temp.columns = [f'{flag}_name', f'{flag}_last_name',
                            f'{flag}_middle_name', f'{flag}_first_name']
            names_df = names_df.merge(temp, how='left', on=[f'{flag}_name'])
        # similar score by element
        for element_name in tqdm(['last_name', 'middle_name', 'first_name']):
            # split data to compare
            condition_compare = names_df[f'raw_{element_name}'].notna(
            ) & names_df[f'best_{element_name}'].notna()
            compare_names_df = names_df[condition_compare].copy()
            not_compare_names_df = names_df[~condition_compare].copy()
            # compare raw with best
            compare_names_df[f'similar_{element_name}'] = compare_names_df[f'raw_{element_name}'].apply(
                unidecode) == compare_names_df[f'best_{element_name}'].apply(unidecode)
            compare_names_df[f'similar_{element_name}'] = compare_names_df[f'similar_{element_name}'].astype(
                int)
            not_compare_names_df[f'similar_{element_name}'] = 1
            # concat
            names_df = pd.concat(
                [compare_names_df, not_compare_names_df], ignore_index=True)
        weights = [0.25, 0.25, 0.5]
        names_df['similarity_score'] = weights[0]*names_df['similar_last_name'] + weights[1] * \
            names_df['similar_middle_name'] + \
            weights[2]*names_df['similar_first_name']
        print(">> similarity_score")
        # Postprocess
        pre_names_df = names_df[['group_id', 'raw_name',
                                 'best_name', 'similarity_score']].copy()
        pre_names_df[key_col] = pre_names_df['group_id'].str.split('-').str[0]
        pre_names_df = pre_names_df.drop(columns=['group_id'])
        pre_names_df = pd.concat(
            [pre_names_df, skip_names_df], ignore_index=True)
        pre_names_df.loc[pre_names_df['best_name'].isna(
        ), 'best_name'] = pre_names_df['raw_name']
        pre_names_df.loc[pre_names_df['similarity_score'].isna(),
                         'similarity_score'] = 1
        print(">> Postprocess")
        # Merge
        pre_names_n = raw_names_n.merge(
            pre_names_df, how='left', on=[key_col, 'raw_name'])
        pre_names_n.loc[pre_names_n['best_name'].isna(
        ), 'best_name'] = pre_names_n['raw_name']
        pre_names_n.loc[pre_names_n['similarity_score'].isna(),
                        'similarity_score'] = 1
        # Return
        pre_names_n = pre_names_n.rename(columns={'raw_name': name_col})

        unify_time = time()-start_time
        print(f"Unify runs in {unify_time/60} mins")

        return pre_names_n



from typing import Tuple

import pandas as pd
from unidecode import unidecode
from tqdm import tqdm

# CONFIG
tqdm.pandas()


def find_match_word(base_word: str, name_dict_df: pd.DataFrame) -> str:
    """
    Find the word in the dictionary if not return the word

    Parameters
    ----------
    base_word : str
        The original word to find inside the dictionary
    name_dict_df : pd.DataFrame
        The DF dictionary containing the words and their diacritics ref

    Returns
    -------
    str
        The returned diacritics version or original word if not have
    """
    try:
        return name_dict_df.loc[name_dict_df['without_accent'] == base_word]['with_accent'].values[0]
    except:
        return base_word


def rule_base_middlename(middlename: str,
                         base_middlename: str,
                         name_dict_df: pd.DataFrame) -> str:
    """
    Rule-base to replace middlename when it is changed by the prediction of the model

    Parameters
    ----------
    middlename : str
        The middlename predicted by the model
    base_middlename : str
        The original middlename when input to model
    name_dict_df : pd.DataFrame
        The DF dictionary containing the middlename and their diacritics ref

    Returns
    -------
    str
        The returned rule-based middlename after being preprocessed
    """

    middle_words = middlename.split()
    base_middle_words = [
        unidecode(word) for word in base_middlename.split() if unidecode(word) != '']
    de_middle_words = [unidecode(word) for word in middle_words]

    try:
        final_middle_words = base_middle_words
        track_idx = list(range(len(base_middle_words)))
        for de_idx, de_word in enumerate(de_middle_words):
            if de_word in base_middle_words:
                if de_word == base_middle_words[de_idx]:
                    final_middle_words[de_idx] = middle_words[de_idx]
                else:
                    base_idx = base_middle_words.index(de_word)
                    final_middle_words[base_idx] = middle_words[de_idx]

                track_idx.remove(de_idx)
    except:
        print(base_middlename, middlename)
        return base_middlename

    # All track are visited
    if not track_idx:
        return ' '.join(final_middle_words)

    # Some word in base is not visited
    for idx in track_idx:
        final_middle_words[idx] = find_match_word(
            base_middle_words[idx], name_dict_df)
    return ' '.join(final_middle_words)


def rule_base_word(word: str, word_base: str, name_dict_df: pd.DataFrame) -> str:
    """
    Apply rule-base to one word

    Parameters
    ----------
    word : str
        The predicted word
    word_base : str
        The original input word
    name_dict_df : pd.DataFrame
        The dictionary DF to track for name

    Returns
    -------
    str
        The best name found in the dictionary
    """
    de_word = unidecode(word)
    if de_word != word_base:
        return find_match_word(word_base, name_dict_df)
    return word


def rule_base_name(name: str, base_name: str, name_dicts: Tuple) -> str:
    """
    Applying rule-based for a specific name with the dictionary of names

    Parameters
    ----------
    name : str
        The predicted name
    base_name : str
        The input name
    name_dicts : Tuple
        The dictionaries to make the rulebase

    Returns
    -------
    str
        Name after go through rule-based postprocessing
    """
    # extract name_dicts
    firstname_dict_df, middlename_dict_df, lastname_dict_df = name_dicts

    # split first, middle, last name
    if base_name is None or len(base_name.split()) == 0:
        return base_name

    firstname = name.split()[-1]
    base_firstname = base_name.split()[-1]
    middlename = ' '.join(name.split()[1:-1])
    base_middlename = ' '.join(base_name.split()[1:-1])
    lastname = name.split()[0]
    base_lastname = base_name.split()[0]

    # take firstname when 1 word
    if firstname == lastname and base_firstname == base_lastname and len(name.split()) == 1:
        return rule_base_word(firstname, base_firstname, firstname_dict_df)

    # applying rule-base
    rule_firstname = rule_base_word(
        firstname, base_firstname, firstname_dict_df)
    rule_middlename = rule_base_middlename(
        middlename, base_middlename, middlename_dict_df)
    rule_lastname = rule_base_word(lastname, base_lastname, lastname_dict_df)

    # joining categories to make full name
    if rule_middlename == '':
        fullname = ' '.join([rule_lastname, rule_firstname])
    else:
        fullname = ' '.join([rule_lastname, rule_middlename, rule_firstname])
    fullname = fullname.replace(r'\s+', ' ').strip()
    return fullname



"""
Module to contains architecture of LSTM
"""

import pickle5 as pickle

import pandas as pd
import tensorflow as tf
from tensorflow.keras.preprocessing.sequence import pad_sequences

from preprocessing_pgp.name.const import GENDER_MODEL_PATH
from preprocessing_pgp.utils import suppress_warnings

suppress_warnings()


class GenderModel:
    def __init__(
        self,
        tokenizer_path: str,
        model_path: str = None
    ):
        self.tokenizer = self.__load_tokenizer(tokenizer_path)
        self.max_len = 5
        self.embed_size = 128
        self.model = self.__load_model(model_path)

    def __load_tokenizer(
        self,
        tokenizer_path: str
    ):
        with open(tokenizer_path, 'rb') as handle:
            tokenizer = pickle.load(handle)

        return tokenizer

    def __load_model(
        self,
        model_path: str = None
    ):
        if model_path is None:
            return

        model = self.build_model()
        model.load_weights(model_path)
        return model

    def build_model(self):
        inputs = tf.keras.layers.Input(shape=(self.max_len, ), name='Input')
        embed_inputs = tf.keras.layers.Embedding(len(self.tokenizer.word_index) + 1,
                                                 self.embed_size,
                                                 name='Embedding')(inputs)
        # Main architecture
        x = tf.keras.layers.LSTM(
            units=32, name='LSTM', return_sequences=True)(embed_inputs)
        x = tf.keras.layers.Dropout(0.5, name='LSTM_dropout')(x)
        x = tf.keras.layers.GlobalMaxPool1D(name='global_max_pool')(x)
        x = tf.keras.layers.Dropout(0.4, name='max_pool_dropout')(x)
        x = tf.keras.layers.Dense(4, activation='relu', name='dense')(x)
        x = tf.keras.layers.Dropout(0.3, name='dense_dropout')(x)

        outputs = tf.keras.layers.Dense(
            1, activation='sigmoid', name='Output')(x)

        model = tf.keras.Model(
            inputs=inputs, outputs=outputs, name='Accented_Model')

        model.compile(
            loss=tf.keras.losses.BinaryCrossentropy(from_logits=False),
            optimizer=tf.keras.optimizers.Adam(learning_rate=1e-3),
            metrics=['accuracy']
        )

        return model

    def predict_gender(
        self,
        names
    ):
        if self.model is None:
            return None
        encode_names = self.tokenizer.texts_to_sequences(names)
        padded_encode_names =\
            pad_sequences(
                encode_names,
                maxlen=self.max_len,
                padding='post'
            )

        genders = tf.cast(
            self.model.predict(padded_encode_names, verbose=0) > 0.5,
            dtype=tf.float32
        )
        return genders


def predict_gender_from_name(
    data: pd.DataFrame,
    name_col: str = 'name'
) -> pd.DataFrame:
    """
    Load model and predict gender from name data

    Parameters
    ----------
    data : pd.DataFrame
        The data contains customer name records
    name_col : str, optional
        The column name of the data that hold name records, by default 'name'

    Returns
    -------
    pd.DataFrame
        Data with additional columns:
        * `gender_predict`: Gender predicted from input names
    """
    if data.empty:
        return data

    gender_model = GenderModel(
        f'{GENDER_MODEL_PATH}/lstm/tokenizer.pkl',
        f'{GENDER_MODEL_PATH}/lstm/gender_lstm.h5'
    )

    data['gender_predict'] = gender_model.predict_gender(
        data[name_col].values
    )
    data['gender_predict'] = data['gender_predict'].map({
        0: 'F',
        1: 'M'
    })

    return data



"""
Module contains package utils
"""
import os
import logging
import multiprocessing as mp
from functools import partial
from typing import (
    Callable,
    List,
    Union,
    Tuple
)

import pandas as pd
import numpy as np
from unidecode import unidecode
from tqdm import tqdm

from preprocessing_pgp.const import (
    N_PROCESSES,
    DICT_TRASH_STRING
)


tqdm.pandas()


def suppress_warnings():
    """
    Function to suppress all possible warnings
    """
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    logging.getLogger('tensorflow').setLevel(logging.ERROR)


def sentence_length(sentence: str) -> int:
    """
    Return the number of words in the sentence
    """
    return len(sentence.split())


def sep_display(sep: str = "\n") -> None:
    """
    Separator for output std
    """
    print(sep)


def is_empty_dataframe(data: pd.DataFrame) -> bool:
    """
    Check whether the dataframe is empty or not
    """
    return data.shape[0] == 0


def replace_trash_string(
    data: pd.DataFrame,
    replace_col: str
) -> pd.DataFrame:
    """
    Replace all trash value with None

    Parameters
    ----------
    data : pd.DataFrame
        The original data
    replace_col : str
        The column that apply the replacement

    Returns
    -------
    pd.DataFrame
        The clean data after replace None
    """
    return data[replace_col].replace(DICT_TRASH_STRING, regex=False)


def extract_null_values(
    data: pd.DataFrame,
    by_col: str
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Extracting NULL values from specific DataFrame

    Parameters
    ----------
    data : pd.DataFrame
        Basic DataFrame
    by_col : str
        Column to separate Null values

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        Tuple of Non-Null DataFrame and Null DataFrame
    """

    null_data = data[data[by_col].isna()].copy()

    non_null_data = data.dropna(subset=[by_col]).copy()

    return non_null_data, null_data


def apply_multi_process(
    func: Callable,
    series: Union[pd.Series, str, np.ndarray],
    n_cores: int = N_PROCESSES
) -> List:
    """
    Process multi-processing on every items of series with provided func

    Parameters
    ----------
    func : Callable
        Function to traverse through series,
        must have 1 input and 1 output
    series : Optional[pd.Series]
        Any series | np.Array() | list
    n_cores : int
        The number of cores used to run parallel, by default half the cores will be used

    Returns
    -------
    List
        List of elements returned after apply the function
    """

    with mp.Pool(n_cores) as pool:
        output = list(pool.imap(func, series))

    return output


def parallelize_dataframe(
    data: pd.DataFrame,
    func: Callable,
    n_cores: int = N_PROCESSES,
    **kwargs
) -> pd.DataFrame:
    """
    Multi-processing on dataframe with provided function and additional function arguments

    Parameters
    ----------
    data : pd.DataFrame
        Any dataframe
    func : Callable
        Function to traverse through separate part of dataframe,
        input must contains the `dataframe` as the required argument
    n_cores : int
        The number of cores used to run parallel, by default half the cores will be used
    **kwargs
        Additional arguments for the function

    Returns
    -------
    pd.DataFrame
        Fully processed dataframe
    """

    if n_cores == 1:
        final_data = partial(func, **kwargs)(data)
    else:
        sub_data = np.array_split(data, n_cores)

        with mp.Pool(n_cores) as pool:
            final_data = pd.concat(pool.map(partial(func, **kwargs), sub_data))

    return final_data


def apply_progress_bar(
    func: Callable,
    series: pd.Series
) -> pd.Series:
    """
    Process apply with progress bar on every items of series with provided func

    Parameters
    ----------
    func : Callable
        Function to traverse through series, must have 1 input and 1 output
    series : pd.Series
        Any series of type pandas Series

    Returns
    -------
    pd.Series
        Series of elements returned after apply the function
    """

    return series.progress_apply(func)


def remove_non_accent_names(
    names_df: pd.DataFrame,
    name_col='name',
    remove_single_name=True
) -> pd.DataFrame:
    """
    Remove non accent names inside the DF

    Parameters
    ----------
    names_df : pd.DataFrame
        The original names DF
    name_col : str, optional
        The column containing the data of names, by default 'name'
    remove_single_name : bool, optional
        Whether to remove a single word name, by default True

    Returns
    -------
    pd.DataFrame
        The clean final DF without any non_accent name
    """
    print("Decoding names...")
    names = names_df[name_col].copy()
    de_names = names.progress_apply(unidecode)

    with_accent_mask = names != de_names

    clean_names = names[with_accent_mask]
    clean_de_names = de_names[with_accent_mask]

    if not remove_single_name:
        len_name = names.apply(lambda name: len(name.split()))
        one_word_mask = len_name == 1
        clean_names = names[with_accent_mask | one_word_mask]
        clean_de_names = de_names[with_accent_mask | one_word_mask]

    clean_names_df = pd.DataFrame({
        'without_accent': clean_de_names,
        'with_accent': clean_names
    })

    without_accent_names_df = names_df[~with_accent_mask].copy()

    return clean_names_df, without_accent_names_df



"""
Module to extract name from email using rule-based
"""
import math

import pandas as pd

from preprocessing_pgp.email.extractors.const import FULLNAME_DICT
from preprocessing_pgp.email.utils import (
    clean_email_name,
)
from preprocessing_pgp.name.type.extractor import process_extract_name_type
from preprocessing_pgp.name.gender.predict_gender import process_predict_gender
from preprocessing_pgp.name.enrich_name import process_enrich

pd.options.mode.chained_assignment = None


class ZipfName:
    """
    Class contains necessary calculation based on Zipfian distribution
    """

    def __init__(self, vocab):
        self.vocab = vocab
        self.vocab_size = len(self.vocab)
        self.__build_cost_dict()

    def __build_cost_dict(self):
        self.word_cost = dict(
            (k, math.log((i+1)*math.log(self.vocab_size)))
            for i, k in enumerate(self.vocab)
        )
        self.max_word_length = max(len(word) for word in self.vocab)

    def extract_component(self, in_str: str) -> str:
        """
        Use dynamic programming to infer the location of spaces in a normed string,
        then return the minimal cost string -- with spaces

        Parameters
        ----------
        in_str : str
            The input normed string

        Returns
        -------
        str
            The returned results of space separated string using Zipfian's law
        """

        cost = [0]

        def get_best_match(str_len: int) -> int:
            """
            Return the best match cost given the len of the string
            """
            start_idx = max(0, str_len-self.max_word_length)
            candidates = enumerate(reversed(
                cost[start_idx:str_len]
            ))

            return min(
                [(c+self.word_cost.get(in_str[str_len-k-1:str_len], 9e999), k+1)
                 for k, c in candidates],
                key=lambda x: x[0]
            )

        # Build the cost array
        in_length = len(in_str)
        for str_len in range(1, in_length+1):
            best_cost, _ = get_best_match(str_len)
            cost.append(best_cost)

        # Backtrack to recover the minimal-cost string
        final_str = []
        str_len = in_length
        while str_len > 0:
            best_cost, k = get_best_match(str_len)
            assert best_cost == cost[str_len], "Wrong logic of calculating cost for string"
            final_str.append(in_str[str_len-k:str_len])
            str_len -= k

        return ' '.join(reversed(final_str))


class EmailNameExtractor:
    """
    Class contains logic to extract username from email
    """

    def __init__(self):
        self.zipf_name = ZipfName(FULLNAME_DICT['name_norm'].tolist())
        self.name_norm_dict =\
            FULLNAME_DICT.set_index('name_norm')\
            .to_dict()['name']

    # ? EXTRACTION COMPONENT OF NAMES
    def _get_username_candidate(
        self,
        data: pd.DataFrame,
        email_name_col: str = 'email_name'
    ) -> pd.DataFrame:
        """
        Extracting username candidate from email's name using ZipF rule

        Parameters
        ----------
        data : pd.DataFrame
            The original dataframe
        email_name_col : str, optional
            The column name contains email_name, by default 'email_name'

        Returns
        -------
        pd.DataFrame
            Component data for username candidate
        """
        data[
            'username_candidate'
        ] = data[
            email_name_col
        ].apply(self.zipf_name.extract_component)

        candidate_data = data['username_candidate'].astype(str)\
            .str.split(' ', expand=True)

        return candidate_data

    def _get_username_from_candidate(
        self,
        data: pd.DataFrame,
        candidate_data: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        Combining candidates of name to make a complete name

        Parameters
        ----------
        data : pd.DataFrame
            Original data
        candidate_data : pd.DataFrame
            Candidate data for name

        Returns
        -------
        pd.DataFrame
            Original data with additional column to get the extracted username
            * `username_extracted` : The extracted name from email
        """

        data['username_extracted'] = ''

        for col in candidate_data.columns:
            # * Whether the candidate is good
            good_name_mask =\
                (candidate_data[col].str.len() > 1) &\
                (candidate_data[col].isin(FULLNAME_DICT['name_norm']))

            # * Translate good candidate to their best name
            candidate_data.loc[
                good_name_mask,
                col
            ] = candidate_data.loc[
                good_name_mask,
                col
            ].map(self.name_norm_dict)

            # * Combining to previous candidates
            data.loc[
                good_name_mask,
                'username_extracted'
            ] = data.loc[
                good_name_mask,
                'username_extracted'
            ] + ' ' + candidate_data.loc[
                good_name_mask,
                col
            ]

        # * Cleansing the spare spaces
        data['username_extracted'] =\
            data['username_extracted'].str.strip()

        return data

    def extract_username(
        self,
        data: pd.DataFrame,
        email_name_col: str = 'email_name',
    ) -> pd.DataFrame:
        """
        Generate username from email data

        Parameters
        ----------
        data : pd.DataFrame
            Data containing the email
        email_name_col : str, optional
            Name of email column from data, by default 'email_name'

        Returns
        -------
        pd.DataFrame
            Data with additional columns:
            * `username_extracted` : The username extracted from email
            * `gender_extracted` : The gender predicted from the extracted username
        """

        extracted_data = data
        orig_cols = extracted_data.columns

        # * Clean the data
        extracted_data = clean_email_name(
            extracted_data,
            email_name_col
        )

        # * Extracting the customer type -- only use 'customer'
        extracted_data = process_extract_name_type(
            extracted_data,
            name_col=f'cleaned_{email_name_col}',
            n_cores=1,
            logging_info=False
        )
        extracted_data[f'cleaned_{email_name_col}'] =\
            extracted_data[f'cleaned_{email_name_col}'].str.lower()

        # * Only proceed clean, 'customer' name
        proceed_mask =\
            (extracted_data[f'cleaned_{email_name_col}'].notna()) &\
            (extracted_data['customer_type'] == 'customer')
        proceed_data = extracted_data[proceed_mask]
        ignored_data = extracted_data[~proceed_mask]

        # * Get email's name candidate
        name_candidate = self._get_username_candidate(
            proceed_data,
            email_name_col=f'cleaned_{email_name_col}'
        )

        # * Extracting username from email
        proceed_data = self._get_username_from_candidate(
            proceed_data,
            name_candidate
        )

        # * Enrich new names
        proceed_data.drop(columns=['customer_type'], inplace=True)
        proceed_data = process_enrich(
            proceed_data,
            name_col='username_extracted',
            n_cores=1,
            logging_info=False
        )

        # * Predict gender from extracted username
        proceed_data = process_predict_gender(
            proceed_data,
            name_col='final',
            n_cores=1,
            logging_info=False
        )
        proceed_data.rename(columns={
            'gender_predict': 'gender_extracted',
            'final': 'enrich_name'
        }, inplace=True)

        # * Combine to get final data
        final_data = pd.concat([
            proceed_data,
            ignored_data
        ])

        return final_data[[
            *orig_cols,
            f'cleaned_{email_name_col}',
            'customer_type',
            'username_extracted',
            'enrich_name',
            'gender_extracted'
        ]]



"""
Module to format the accent typing to old typing
"""
import regex as re
import unidecode as decode


# REFORMATING ACCENT
bang_nguyen_am = [['a', 'à', 'á', 'ả', 'ã', 'ạ', 'a'],
                  ['ă', 'ằ', 'ắ', 'ẳ', 'ẵ', 'ặ', 'aw'],
                  ['â', 'ầ', 'ấ', 'ẩ', 'ẫ', 'ậ', 'aa'],
                  ['e', 'è', 'é', 'ẻ', 'ẽ', 'ẹ', 'e'],
                  ['ê', 'ề', 'ế', 'ể', 'ễ', 'ệ', 'ee'],
                  ['i', 'ì', 'í', 'ỉ', 'ĩ', 'ị', 'i'],
                  ['o', 'ò', 'ó', 'ỏ', 'õ', 'ọ', 'o'],
                  ['ô', 'ồ', 'ố', 'ổ', 'ỗ', 'ộ', 'oo'],
                  ['ơ', 'ờ', 'ớ', 'ở', 'ỡ', 'ợ', 'ow'],
                  ['u', 'ù', 'ú', 'ủ', 'ũ', 'ụ', 'u'],
                  ['ư', 'ừ', 'ứ', 'ử', 'ữ', 'ự', 'uw'],
                  ['y', 'ỳ', 'ý', 'ỷ', 'ỹ', 'ỵ', 'y']]
nguyen_am_to_ids = {}


def is_valid_vietnam_word(word: str) -> bool:
    """
    Return `True` if the input word is vietnamese word, else `False`

    Parameters
    ----------
    word : str
        The input word to check for correct Vietnamese syntax

    Returns
    -------
    bool
        Whether the input word is Vietnamese
    """
    chars = list(word)
    nguyen_am_index = -1
    for index, char in enumerate(chars):
        x, _ = nguyen_am_to_ids.get(char, (-1, -1))
        if x != -1:
            if nguyen_am_index == -1:
                nguyen_am_index = index
            else:
                if index - nguyen_am_index != 1:
                    return False
                nguyen_am_index = index
    return True


def reformat_vi_word_accent(word: str) -> str:
    """
    Reformat the word to correct accent typing if it is Vietnamese word, otherwise return the word and do nothing

    Parameters
    ----------
    word : str
        The input word to transform to correct typing

    Returns
    -------
    str
        The correct typing of the word if it is Vietnamese, otherwise the original input
    """
    if not is_valid_vietnam_word(word):
        return word

    chars = list(word)
    dau_cau = 0
    nguyen_am_index = []
    qu_or_gi = False
    for index, char in enumerate(chars):
        x, y = nguyen_am_to_ids.get(char, (-1, -1))
        if x == -1:
            continue
        elif x == 9:  # check qu
            if index != 0 and chars[index - 1] == 'q':
                chars[index] = 'u'
                qu_or_gi = True
        elif x == 5:  # check gi
            if index != 0 and chars[index - 1] == 'g':
                chars[index] = 'i'
                qu_or_gi = True
        if y != 0:
            dau_cau = y
            chars[index] = bang_nguyen_am[x][0]
        if not qu_or_gi or index != 1:
            nguyen_am_index.append(index)
    if len(nguyen_am_index) < 2:
        if qu_or_gi:
            if len(chars) == 2:
                x, y = nguyen_am_to_ids.get(chars[1])
                chars[1] = bang_nguyen_am[x][dau_cau]
            else:
                x, y = nguyen_am_to_ids.get(chars[2], (-1, -1))
                if x != -1:
                    chars[2] = bang_nguyen_am[x][dau_cau]
                else:
                    chars[1] = bang_nguyen_am[5][dau_cau] if chars[1] == 'i' else bang_nguyen_am[9][dau_cau]
            return ''.join(chars)
        return word

    for index in nguyen_am_index:
        x, y = nguyen_am_to_ids[chars[index]]
        if x == 4 or x == 8:  # ê, ơ
            chars[index] = bang_nguyen_am[x][dau_cau]
            # for index2 in nguyen_am_index:
            #     if index2 != index:
            #         x, y = nguyen_am_to_ids[chars[index]]
            #         chars[index2] = bang_nguyen_am[x][0]
            return ''.join(chars)

    if len(nguyen_am_index) == 2:
        if nguyen_am_index[-1] == len(chars) - 1:
            x, y = nguyen_am_to_ids[chars[nguyen_am_index[0]]]
            chars[nguyen_am_index[0]] = bang_nguyen_am[x][dau_cau]
            # x, y = nguyen_am_to_ids[chars[nguyen_am_index[1]]]
            # chars[nguyen_am_index[1]] = bang_nguyen_am[x][0]
        else:
            # x, y = nguyen_am_to_ids[chars[nguyen_am_index[0]]]
            # chars[nguyen_am_index[0]] = bang_nguyen_am[x][0]
            x, y = nguyen_am_to_ids[chars[nguyen_am_index[1]]]
            chars[nguyen_am_index[1]] = bang_nguyen_am[x][dau_cau]
    else:
        # x, y = nguyen_am_to_ids[chars[nguyen_am_index[0]]]
        # chars[nguyen_am_index[0]] = bang_nguyen_am[x][0]
        x, y = nguyen_am_to_ids[chars[nguyen_am_index[1]]]
        chars[nguyen_am_index[1]] = bang_nguyen_am[x][dau_cau]
        # x, y = nguyen_am_to_ids[chars[nguyen_am_index[2]]]
        # chars[nguyen_am_index[2]] = bang_nguyen_am[x][0]
    return ''.join(chars)


def reformat_vi_sentence_accent(sentence):
    """
    Change the Vietnamese accent typing to the old format

    Parameters
    ----------
    sentence : str
        Input sentence to be reformatted
    """

    sentence = sentence.lower()
    words = sentence.split(' ')

    for index, word in enumerate(words):
        cw = re.sub(r'(^\p{P}*)([p{L}.]*\p{L}+)(\p{P}*$)',
                    r'\1/\2/\3', word).split('/')
        # print(cw)

        if len(cw) == 3:
            cw[1] = reformat_vi_word_accent(cw[1])

        words[index] = ''.join(cw)

    return ' '.join(words)


def remove_accent_typing(sentence):
    """
    Remove the accent typing of Vietnamese sentence

    Parameters
    ----------
    sentence : str
        Input sentence to be reformatted
    """

    return decode.unidecode(sentence)



import os

N_PROCESSES = os.cpu_count() // 2

DICT_TRASH_STRING = {
    '': None,
    'Nan': None,
    'nan': None,
    'None': None,
    'none': None,
    'Null': None,
    'null': None,
    '""': None
}



"""
Module to predict for gender from name
"""
from time import time

import pandas as pd

from preprocessing_pgp.name.type.extractor import process_extract_name_type
from preprocessing_pgp.name.preprocess import preprocess_df
from preprocessing_pgp.name.model.lstm import predict_gender_from_name
from preprocessing_pgp.utils import (
    parallelize_dataframe,
)


def process_predict_gender(
    data: pd.DataFrame,
    name_col: str = 'name',
    n_cores: int = 1,
    logging_info: bool = True
) -> pd.DataFrame:
    """
    Process predict gender from names

    Parameters
    ----------
    data : pd.DataFrame
        The data contains the column with username
    name_col : str, optional
        The column name contains records of username, by default 'name'
    n_cores : int, optional
        The number of core to process, by default 1
    logging_info : bool, optional
        Whether to log info while running, by default True

    Returns
    -------
    pd.DataFrame
        Data with additional column:
        * `gender_predict`: predicted gender from name
    """
    name_data = data[[name_col]].dropna()
    nan_data = data[data[name_col].isna()][[name_col]]
    orig_cols = data.columns

    # * Clean name data
    if logging_info:
        print(">>> Cleansing name: ", end='')
    start_time = time()
    cleaned_name_data = parallelize_dataframe(
        name_data,
        preprocess_df,
        n_cores=n_cores,
        name_col=name_col
    )
    clean_time = time() - start_time
    if logging_info:
        print(f"{int(clean_time)//60}m{int(clean_time)%60}s")

    # * Get customer type
    cleaned_name_data = process_extract_name_type(
        cleaned_name_data,
        name_col=name_col,
        n_cores=n_cores,
        logging_info=False
    )

    # * Only predict for customer's name
    customer_mask = cleaned_name_data['customer_type'] == 'customer'
    customer_name_data = cleaned_name_data[customer_mask]
    non_customer_name_data = cleaned_name_data[~customer_mask]

    # * Predict gender
    if logging_info:
        print(">>> Predicting gender: ", end='')
    start_time = time()
    predicted_name_data = parallelize_dataframe(
        customer_name_data,
        predict_gender_from_name,
        n_cores=n_cores,
        name_col=name_col
    )
    predict_time = time() - start_time
    if logging_info:
        print(f"{int(predict_time)//60}m{int(predict_time)%60}s")

    # * Concat to generate final data
    new_cols = ['gender_predict']
    nan_data[new_cols] = None
    final_data = pd.concat(
        [predicted_name_data, nan_data, non_customer_name_data])
    final_data = pd.concat(
        [data[orig_cols], final_data[new_cols]], axis=1)

    return final_data



# __main__.py

import sys
import argparse

import pandas as pd


def main():
    print("Project is on the way to update this feature...,\nThank you all for your support")
    return


if __name__ == '__main__':
    main()



import re
from typing import List


def number_pad_replace(match: re.Match) -> str:
    """
    Replacement function for removing padding in number string

    Parameters
    ----------
    match : re.Match
        match object received by regex

    Returns
    -------
    str
        return string without padding
    """

    number = int(match.group(1))

    return format(number, '01d')


def is_empty_string(string: str) -> bool:
    """
    Check if the string is empty or not
    """
    return string == ''


def create_dependent_query(*dependents) -> str:
    """
    Make a `pd.DataFrame` query to match all dependents

    * `dependents` should be strings of comparison query
    """
    query = ''
    if len(dependents) > 0:
        for dep in dependents:
            if len(dep) > 0:
                query = f'{query} & {dep}'\
                    if query != ''\
                    else dep

    return query


def flatten_list(lst: List) -> List:
    """
    Helper function to flatten the list of list into simpler list
    """
    flat_list = [item for sublist in lst for item in sublist]

    return flat_list


def remove_substr(string: str, substr: str) -> str:
    """
    Remove sub-string from original `string` and return the modified string

    Parameters
    ----------
    string : str
        The string to remove sub-string
    substr : str
        The sub-string in string to be removed

    Returns
    -------
    str
        The modified string without sub-string,
        if not having any sub-str found or substring is empty
        return the origin string
    """
    if None in [string, substr]:
        return None

    start_sub_idx = string.rfind(substr)

    if start_sub_idx == -1 or len(substr) == 0:  # * Failure if s
        return string

    end_sub_idx = start_sub_idx + len(substr)

    ret_str = string[:start_sub_idx] + string[end_sub_idx:]

    return ret_str.strip()



import string
import re
import os
import sys
from string import punctuation

import pandas as pd

from preprocessing_pgp.accent_typing_formatter import reformat_vi_sentence_accent
from preprocessing_pgp.name.unicode_converter import minimal_convert_unicode
from preprocessing_pgp.name.extract_human import replace_non_human_reg
from preprocessing_pgp.name.split_name import NameProcess

_dir = "/".join(os.path.split(os.getcwd()))
if _dir not in sys.path:
    sys.path.append(_dir)


def remove_special_chars(sentence: str) -> str:
    """
    Removing special characters from sentence

    Parameters
    ----------
    sentence : str
        The input sentence can contains many special characters and alpha characters

    Returns
    -------
    str
        The sentence contains only alpha characters
    """
    return sentence.translate(str.maketrans('', '', punctuation))


def remove_non_word(sentence: str) -> str:
    """
    Removing all non-word character from sentence

    Parameters
    ----------
    sentence : str
        The input sentence can contains symbols, special non-utf8 characters

    Returns
    -------
    str
        The sentence not contain any special character
    """
    clean_sentence = re.sub(r'[^\w]', ' ', sentence)
    return clean_sentence


def remove_spare_spaces(sentence: str) -> str:
    """
    Removing spare spaces inside a sentence

    Parameters
    ----------
    sentence : str
        Sentence to clean spare spaces

    Returns
    -------
    str
        Cleaned sentence
    """

    # Remove spaces in between
    sentence = re.sub(' +', ' ', sentence)
    # Remove additional leading & trailing spaces
    sentence = sentence.strip()

    return sentence


def format_caps_word(sentence: str) -> str:
    """
    Format the sentence into capital format

    Parameters
    ----------
    sentence : str
        The input sentences

    Returns
    -------
    str
        Capitalized sentence of the input sentence
    """
    caps_sen = sentence.lower()
    caps_sen = string.capwords(caps_sen)

    return caps_sen


def basic_preprocess_name(name: str) -> str:
    """
    Preprocess names based on these steps:

        1. Remove spare spaces
        2. Format name into capitalized
        3. Change name to Unicode compressed format
        4. Change name to same old accent typing format
        5. Re-capitalized the word

    Parameters
    ----------
    name : str
        The input raw name

    Returns
    -------
    str
        The preprocessed name
    """
    # Remove Special Characters
    clean_name = remove_special_chars(name)

    # Remove all non-word characters
    clean_name = remove_non_word(clean_name)

    # Remove Spare Spaces
    clean_name = remove_spare_spaces(clean_name)

    # Format Caps
    caps_name = format_caps_word(clean_name)

    # Change to same VN charset -> Unicode compressed
    unicode_clean_name = minimal_convert_unicode(caps_name)

    # Change to same accent typing -> old type
    old_unicode_clean_name = reformat_vi_sentence_accent(unicode_clean_name)

    old_unicode_clean_name = format_caps_word(old_unicode_clean_name)

    return old_unicode_clean_name


def clean_name_cdp(name: str) -> str:
    """
    Specific function to clean name from customer profile

    Parameters
    ----------
    name : str
        The name of the customer

    Returns
    -------
    str
        The clean name of the customer
    """
    if name is None:
        return None
    clean_name = basic_preprocess_name(name)
    clean_name = replace_non_human_reg(name)
    return clean_name


def preprocess_df(
    data: pd.DataFrame,
    # human_extractor: HumanNameExtractor,
    name_col: str = 'name',
    clean_name: bool = True
    # extract_human: bool = False,
    # multiprocessing: bool = False,
    # n_cpu: int = None
) -> pd.DataFrame:
    """
    Perform basic preprocessing to names in the input data

    Parameters
    ----------
    df : pd.DataFrame
        The input data containing the columns with name records
    name_col : str, optional
        The column contains the name records, by default 'name'
    clean_name : bool, optional
        Whether to clean the name separated by space, by default True

    Returns
    -------
    pd.DataFrame
        The finalized data with clean names
    """
    # * Filter out columns
    other_cols = data.columns.difference([name_col])

    # * Na names & filter out name col
    na_data = data[data[name_col].isna()][[name_col]]
    cleaned_data = data[data[name_col].notna()][[name_col]]

    # * Cleansing data
    cleaned_data[f'clean_{name_col}'] =\
        cleaned_data[name_col].apply(
            basic_preprocess_name
    )
    if clean_name:
        name_process = NameProcess()
        cleaned_data[f'clean_{name_col}'] =\
            cleaned_data[f'clean_{name_col}'].apply(
                lambda name: name_process.CleanName(name)[0]
        )

    cleaned_data = cleaned_data.drop(columns=[name_col])
    cleaned_data = cleaned_data.rename(columns={
        f'clean_{name_col}': name_col
    })

    # * Concat na data
    final_data = pd.concat([
        cleaned_data,
        na_data
    ])

    # * Concat with original cols
    new_cols = [
        name_col
    ]
    final_data = pd.concat([data[other_cols], final_data[new_cols]], axis=1)

    return final_data



"""
Module contains constants for email information extraction
"""

import os

import pandas as pd

# ? STUDENT FULLNAME
fullname_path = os.path.join(
    os.path.dirname(__file__),
    '../../data/email_info/norm_name_trace_dict.parquet'
)

FULLNAME_DICT = pd.read_parquet(fullname_path)

# ? YOB
dob_path = os.path.join(
    os.path.dirname(__file__),
    '../../data/email_info/dob_dict.parquet'
)
DOB_DICT = pd.read_parquet(dob_path)

DOB_REGEX_DICT = DOB_DICT.set_index('dob_type')['regex'].to_dict()
DOB_NUM_DIGIT_DICT = DOB_DICT.set_index('dob_type')['num_digit'].to_dict()
DOB_FORMAT_DICT = DOB_DICT.set_index('dob_type')['format'].to_dict()

# ? PHONE

VIETNAMESE_PHONE_REGEX = r'((02[0-9]|0|84)[3|5|7|8|9]*[0-9]{8})'



from preprocessing_pgp.phone.const import (
    SUB_PHONE_11NUM,
    SUB_TELEPHONE_10NUM,
    DICT_4_SUB_PHONE,
    DICT_4_SUB_TELEPHONE
)


def convert_mobi_phone(phone: str) -> str:
    """
    Convert old mobiphone format(11 numbers) into new format(10 numbers)

    Parameters
    ----------
    phone : str
        The old format phone

    Returns
    -------
    str
        Newly returned phone with new format
    """
    if phone[:4] in SUB_PHONE_11NUM:
        return DICT_4_SUB_PHONE[phone[:4]] + phone[4:]
    else:
        return None


def convert_phone_region(old_region: str) -> str:
    """
    Convert old region code phone to new region code phone

    Parameters
    ----------
    old_region : str
        Phone from old region

    Returns
    -------
    str
        New phone from new region
    """
    if old_region[:2] in SUB_TELEPHONE_10NUM:
        return DICT_4_SUB_TELEPHONE[old_region[:2]] + old_region[2:]

    if old_region[:3] in SUB_TELEPHONE_10NUM:
        return DICT_4_SUB_TELEPHONE[old_region[:3]] + old_region[3:]

    if old_region[:4] in SUB_TELEPHONE_10NUM:
        return DICT_4_SUB_TELEPHONE[old_region[:4]] + old_region[4:]

    return None



"""
Module to preprocess email before proceed more
"""

import re
import pandas as pd
from preprocessing_pgp.email.const import NAN_EMAIL_LIST


class EmailCleaner:
    """
    Class support cleansing function for Email
    """

    def clean_email(
        self,
        email: str
    ) -> str:
        """
        Process cleaning email:
        1. Remove spaces

        Parameters
        ----------
        email : str
            The input email to clean

        Returns
        -------
        str
            Cleaned email without any spaces
        """

        # Case NaN email
        if email is None or email in NAN_EMAIL_LIST:
            return None

        cleaned_email = email.lower()
        cleaned_email = self._remove_spaces(cleaned_email)

        return cleaned_email

    def _remove_spaces(
        self,
        email: str
    ) -> str:
        """
        Function to remove spaces from email
        """
        cleaned_email = re.sub(' +', '', email)

        return cleaned_email


def clean_email(
    data: pd.DataFrame,
    email_col: str = 'email'
) -> pd.DataFrame:
    """
    Process cleansing email from dataframe
    and created `cleaned_<email_col>` column contains the cleaned email

    Parameters
    ----------
    data : pd.DataFrame
        The input dataframe contains a column with email
    email_col : str, optional
        The name of the column contains email records, by default 'email'

    Returns
    -------
    pd.DataFrame
        The original dataframe with an additional column contains cleaned email
    """
    if data.empty:
        return data

    cleaner = EmailCleaner()

    cleaned_data = data

    cleaned_data[f'cleaned_{email_col}'] =\
        cleaned_data[email_col].apply(
            cleaner.clean_email
    )

    return cleaned_data



"""
Constants for processing name type extraction
"""

import os

import pandas as pd

# ? IMPORTANT PATHS
__NAME_TYPE_PATH = os.path.join(
    os.path.dirname(__file__),
    '../../',
    'data',
    'name_type'
)

_NAME_TYPE_LV1_PATH = os.path.join(
    __NAME_TYPE_PATH,
    'customer_type_lv1.parquet'
)

_NAME_TYPE_LV2_PATH = os.path.join(
    __NAME_TYPE_PATH,
    'customer_type_lv2.parquet'
)

# ? DATA BY LEVELS
LV1_NAME_TYPE = pd.read_parquet(_NAME_TYPE_LV1_PATH)
LV2_NAME_TYPE = pd.read_parquet(_NAME_TYPE_LV2_PATH)
NAME_TYPE_DATA = {
    'lv1': LV1_NAME_TYPE,
    'lv2': LV2_NAME_TYPE
}

# ? NAME TYPE REGEX
LV1_NAME_TYPE_REGEX = LV1_NAME_TYPE.groupby('ctype')['term'].apply(list).to_dict()
LV2_NAME_TYPE_REGEX = LV2_NAME_TYPE.groupby('ctype')['term'].apply(list).to_dict()
NAME_TYPE_REGEX_DATA = {
    'lv1': LV1_NAME_TYPE_REGEX,
    'lv2': LV2_NAME_TYPE_REGEX
}



"""
Module to contains architecture of Naive Bayes SVM -- Deprecated
"""

import string

import pickle
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y, check_is_fitted
from sklearn.linear_model import LogisticRegression
from scipy import sparse

from preprocessing_pgp.name.const import GENDER_MODEL_PATH

class NBSVMClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, **logistic_kws):
        self.logistic_kws = logistic_kws

    def predict(self, x):
        # Verify that model has been fit
        check_is_fitted(self, ['_r', '_clf'])
        return self._clf.predict(x.multiply(self._r))

    def predict_proba(self, x):
        # Verify that model has been fit
        check_is_fitted(self, ['_r', '_clf'])
        return self._clf.predict_proba(x.multiply(self._r))

    def fit(self, x, y):
        # Check that X and y have correct shape
        x, y = check_X_y(x, y, accept_sparse=True)

        def pr(x, y_i, y):
            p = x[y==y_i].sum(0)
            return (p+1) / ((y==y_i).sum()+1)

        self._r = sparse.csr_matrix(np.log(pr(x,1,y) / pr(x,0,y)))
        x_nb = x.multiply(self._r)

        self._clf = LogisticRegression(**self.logistic_kws).fit(x_nb, y)
        return self

def load_model(file_path):
    with open(file_path, 'rb') as pkl_model:
        model = pickle.load(pkl_model)
    return model

def preprocess_fromName(df, name_col='Name'):
    df = df.copy()
    df['CleanName'] = df[name_col]\
        .str.lower()\
        .str.replace(r'\d', ' ', regex=True)\
        .str.replace(rf'[{string.punctuation}]', ' ', regex=True)\
        .str.replace(r'\s+', ' ', regex=True)\
        .str.strip()
    return df

def get_gender_from_accent_name(df, name_col='Name'):
    prep_df = preprocess_fromName(df.copy(), name_col)
    pipeline = load_model(f'{GENDER_MODEL_PATH}/accented/nbsvm_pipeline_v3.pkl')
    predictions = pipeline.predict(prep_df['CleanName'].values)
    predictions = list(map(lambda x: 'M' if x == 1 else 'F', predictions))
    return predictions

def get_gender_from_non_accent_name(df, name_col='Name'):
    prep_df = preprocess_fromName(df.copy(), name_col)
    pipeline = load_model(f'{GENDER_MODEL_PATH}/non_accented/nbsvm_pipeline_v3.pkl')
    predictions = pipeline.predict(prep_df['CleanName'].values)
    predictions = list(map(lambda x: 'M' if x == 1 else 'F', predictions))
    return predictions



"""
File containing code that related to n-level extraction of address
"""

from typing import List, Tuple, Dict
from copy import deepcopy

import pandas as pd
from flashtext import KeywordProcessor

from preprocessing_pgp.address.utils import (
    flatten_list,
    remove_substr,
    create_dependent_query
)
from preprocessing_pgp.address.const import (
    METHOD_REFER_DICT,
    LOCATION_ENRICH_DICT
)


class LevelExtractor:
    """
    Class contains information of extract level and functions to process on each level

    * Level 1: City, Countryside
    * Level 2: District
    * Level 3: Ward
    """

    def __init__(self) -> None:
        self.avail_levels = METHOD_REFER_DICT.keys()
        self.avail_methods = flatten_list(METHOD_REFER_DICT.values())
        self.keyword_refer_dict =\
            dict(zip(
                self.avail_methods,
                [self._generate_keyword_processor(method)
                 for method in self.avail_methods]
            ))

    def _get_level_methods(self, level) -> List:
        """
        _summary_

        Parameters
        ----------
        level : _type_
            _description_

        Returns
        -------
        List
            _description_
        """
        return METHOD_REFER_DICT[level]

    def _get_level_keywords(self, level: int = 1) -> List:
        """
        Function to get all keywords of each specific level

        Parameters
        ----------
        level : int, optional
            The number of level (1-3) to generate method keywords, by default 1

        Returns
        -------
            The output list of keywords for each of the level's methods
        """
        level_kws = []
        for method in self._get_level_methods(level):
            method_kw =\
                self._generate_keyword_processor(method)
            level_kws.append(method_kw)

        return level_kws

    def _generate_keyword_processor(self, method: str) -> KeywordProcessor:
        """
        Function to generate `KeywordProcessor` object
        to the specific `method`
        using the generated dictionary of keywords

        Parameters
        ----------
        method : str
            The method string refer to the dictionary column name

        Returns
        -------
        KeywordProcessor
            `KeywordProcessor` object that contains all the keywords of the method
        """
        unique_method_kws =\
            LOCATION_ENRICH_DICT[method].unique().tolist()

        keyword_processor = KeywordProcessor(case_sensitive=True)
        keyword_processor.add_keywords_from_list(unique_method_kws)

        return keyword_processor

    def _get_match_keyword(
        self,
        query: str,
        kw_processor: KeywordProcessor
    ) -> str:
        """
        Function to get the last keyword found in `query` string

        Parameters
        ----------
        query : str
            The input query string containing the address
        kw_processor : KeywordProcessor
            The keyword processor for specific method in dictionary (norm, abbrev, ...)

        Returns
        -------
        str
            Last match keyword found in query, if not found return None
        """
        found_keywords = kw_processor.extract_keywords(query)

        if len(found_keywords) > 0:
            return found_keywords[-1]

        return None

    def extract_all_levels(self, address: str) -> Tuple[Dict, str]:
        """
        Traverse through all possible level from lowest to highest
        to check and find `best pattern` in each level
        and `return` as a `dictionary of pattern` found at each level with `remaining address`

        Parameters
        ----------
        address : str
            The unified address that has been `lowered` and `unidecode`

        Returns
        -------
        Tuple[Dict, str]
            The output contains:
            * a dictionary of `pattern` found at each level
            * `remained address`
            * a dictionary of `best pattern` found at each level
        """

        remained_address = deepcopy(address)
        found_patterns =\
            dict(zip(
                self.avail_levels,
                [None]*len(self.avail_levels)
            ))
        best_patterns = deepcopy(found_patterns)

        dependents = []
        for level in self.avail_levels:
            (level_pattern,
             remained_address,
             level_method,
             level_best_pattern) =\
                self._extract_by_level(remained_address, level, *dependents)

            found_patterns[level] = level_pattern
            best_patterns[level] = level_best_pattern
            dependents.append((level_pattern, level_method))

        return found_patterns, remained_address, best_patterns

    def __is_query_exist(
        self,
        query: str
    ) -> bool:
        """
        Helper function to check whether the query is possibly found in location data
        """
        n_result = LOCATION_ENRICH_DICT.query(query).shape[0]

        return n_result > 0

    def __make_dependent_query(
        self,
        *dependents
    ) -> str:
        """
        Helper to make search query from dependents
        """
        query = ''
        # Making dependent queries
        if len(dependents) > 0:
            dependent_queries = []
            for d_term, d_method in dependents:
                if d_term is not None:
                    term_query = f'{d_method} == "{d_term}"'
                    dependent_queries.append(term_query)

            query = create_dependent_query(*dependent_queries)

        return query

    def __is_correct_dependent(
        self,
        *dependents
    ) -> bool:
        """
        Helper function to check whether the dependencies are all correct
        """
        query = self.__make_dependent_query(*dependents)

        if query != '':
            return self.__is_query_exist(query)

        return False

    def _extract_by_level(
        self,
        address: str,
        level: int,
        *dependents
    ) -> Tuple[str, str, str, str]:
        """
        Extract address with list of `method`
        at the specific `level`
        to find the best matched pattern

        Parameters
        ----------
        address : str
            The unified address that has been `lowered` and `unidecode`
        level : int
            The level which is currently traversed from

        Returns
        -------
        Tuple[str, str, str, str]
            The output contains:
            * `pattern` found within specific level
            * `remained address`
            * `method` which retrieved the pattern
            * `best pattern` unified by `pattern`
        """
        level_methods = self._get_level_methods(level)

        for method in level_methods:
            pattern_found, remained_address =\
                self._extract_by_method(address, method)

            # Found something then return
            if (pattern_found is not None)\
                    and (len(pattern_found) > 0):
                if self.__is_correct_dependent(*dependents, (pattern_found, method)):
                    best_pattern = self.__trace_match_pattern(
                        pattern_found,
                        method,
                        *dependents
                    )
                    return pattern_found, remained_address, method, best_pattern

        return None, address, None, None

    def _extract_by_method(
        self,
        address: str,
        method: str
    ) -> Tuple[str, str]:
        """
        Extract the address with `keywords` of the specific method

        Parameters
        ----------
        address : str
            The unified address that has been `lowered` and `unidecode`
        method : str
            The method of level which is currently being explored

        Returns
        -------
        Tuple[str, str]
            The output contains:
            * `pattern` found within specific method
            * `remained address`
        """
        method_kw = self.keyword_refer_dict[method]
        match_pattern = self._get_match_keyword(address, method_kw)
        remained_address = remove_substr(address, match_pattern)

        return match_pattern, remained_address

    def __trace_match_pattern(
        self,
        pattern: str,
        method: str,
        *dependents
    ) -> str:
        """
        Helper function to trace back best pattern found
        """
        query = self.__make_dependent_query(*dependents, (pattern, method))

        if not self.__is_query_exist(query):
            return None

        level_col = method[:3]

        best_match = self.__trace_best_match(query, level_col)

        return best_match

    def __trace_best_match(
        self,
        query: str,
        level_col: str
    ) -> str:
        """
        Helper to get the best match consist of:

        1. Exist
        2. Unique found
        """
        found_terms = LOCATION_ENRICH_DICT.query(query)[level_col].unique()

        n_found = found_terms.shape[0]
        if n_found == 0 or n_found > 1:
            return None

        return found_terms[0]


def extract_vi_address_by_level(
    data:  pd.DataFrame,
    address_col: str
) -> pd.DataFrame:
    """
    Function to extract vietnamese address into each specific level found by pattern

    Parameters
    ----------
    data : pd.DataFrame
        Raw data containing the address
    address_col : str
        The raw address column that need cleansing and unifying

    Returns
    -------
    pd.DataFrame
        Final data with new columns:

        * `level_1`: city, countryside found
        * `best_level_1`: beautified city, countryside found
        * `level_2`: district found
        * `best_level_2`: beautified district found
        * `level_3`: ward found
        * `best_level_3`: beautified ward found
        * `remained_address`: the remaining in the address
    """

    extractor = LevelExtractor()

    extracted_data = data

    extracted_results =\
        extracted_data[address_col].apply(
            extractor.extract_all_levels
        )

    for level in extractor.avail_levels:
        extracted_data[f'level_{level}'] =\
            [result[0][level]
             for result in extracted_results]

        extracted_data[f'best_level_{level}'] =\
            [result[-1][level]
             for result in extracted_results]

    extracted_data['remained_address'] =\
        [result[1]
         for result in extracted_results]

    return extracted_data



"""
Module contains constants for email processing
"""

# ? NONE TYPE EMAIL DICT
NAN_EMAIL_LIST = ['nan', 'none', '']

# ? EMAIL REGEX

AT_LEAST_ONE_CHAR_REGEX = r'(?=.*[a-z])'

EMAIL_DOMAIN_REGEX = {
    'gmail': {
        'domains': r'gmail.com([.][a-z]{2})?$',
        'regex': r'^(?!.*google)[a-z][a-z0-9\.]{4,28}[a-z0-9]'
    },
    'yahoo': {
        'domains': r'(yahoo|ymail|myyahoo).com([.][a-z]{2})?$',
        'regex': r'^[a-z][a-z0-9_\.]{2,30}[a-z0-9]'
    },
    'ms': {
        'domains': r'(hotmail|outlook).com([.][a-z]{2})?$',
        'regex': r'^[a-z][a-z0-9-_\.]{0,30}[a-z0-9-_]?'
    },
    'fpt': {
        'domains': r'.*fpt.*$',
        'regex': r'^[a-z0-9][a-z0-9_\.]{2,31}[a-z0-9]'
    }
}

COMMON_EMAIL_REGEX = r'^[a-z0-9][a-z0-9_\.]{2,30}[a-z0-9]@[a-z0-9]{2,}(?:\.[a-z0-9]{2,12}){1,2}$'

EDU_EMAIL_REGEX = r'^[a-z0-9][a-z0-9_.]{4,30}[a-z0-9]'

EDGE_AUTO_EMAIL_REGEX = r'@privaterelay.appleid.com|[0-9a-z]+\_autoemail'

# ? EMAIL CONSTRAINTS
DOMAIN_GROUP_DICT = dict((reg_li['domains'], domain) for domain, reg_li in EMAIL_DOMAIN_REGEX.items())

PRIVATE_EMAIL_DOMAINS = [
    'gmail.com', 'yahoo.com',
    'yahoo.com.vn', 'icloud.com',
    'email.com', 'hotmail.com', 'gmai.com', 'outlook.com']
LEAST_NUM_EMAIL_CHAR = 8
LONGEST_VIETNAMESE_WORD = 'Nghieng'



"""
File containing logic to extract a full address into 3 levels:

* Level 1: City, Countryside
* Level 2: District
* Level 3: Ward
"""
from time import time

import pandas as pd

from preprocessing_pgp.address.loc_process import generate_loc_code
from preprocessing_pgp.address.level_extractor import extract_vi_address_by_level
from preprocessing_pgp.address.preprocess import clean_vi_address
from preprocessing_pgp.utils import (
    parallelize_dataframe,
    extract_null_values
)
from preprocessing_pgp.address.const import AVAIL_LEVELS


def extract_vi_address(
    data: pd.DataFrame,
    address_col: str,
    n_cores: int = 1,
    logging_info: bool = True
) -> pd.DataFrame:
    """
    Extract Vietnamese address by pattern to find 3 levels of address

    Parameters
    ----------
    data : pd.DataFrame
        The input raw data with address column
    address_col : str
        The name of the column containing addresses
    n_cores : int, optional
        The number of cores used to run parallel, by default 1 core will be used

    Returns
    -------
    pd.DataFrame
        The data with additional columns:

        * `cleaned_<address_col>` contains the unified and clean Vietnamese address
        * `level_1`: city, countryside found
        * `best_level_1`: beautified city, countryside found
        * `level_1_code`: code for city
        * `level_2`: district found
        * `best_level_2`: beautified district found
        * `level_2_code`: code for district
        * `level_3`: ward found
        * `best_level_3`: beautified ward found
        * `level_3_code`: code for ward
        * `remained address`: the remaining in the address
    """

    # * Select only address column
    orig_cols = data.columns
    address_data = data[[address_col]]

    # * Removing na addresses
    clean_address_data, na_address_data =\
        extract_null_values(
            address_data,
            by_col=address_col
        )

    # * Cleanse the address
    if logging_info:
        print(">>> Cleansing address: ", end='')
    start_time = time()
    cleaned_data = parallelize_dataframe(
        clean_address_data,
        clean_vi_address,
        n_cores=n_cores,
        address_col=address_col
    )
    clean_time = time() - start_time
    if logging_info:
        print(f"{int(clean_time)//60}m{int(clean_time)%60}s")

    # * Feed the cleansed address to extract the level
    if logging_info:
        print(">>> Extract & Beautify address: ", end='')
    start_time = time()
    extracted_data = parallelize_dataframe(
        cleaned_data,
        extract_vi_address_by_level,
        n_cores=n_cores,
        address_col=f'cleaned_{address_col}'
    )
    extract_time = time() - start_time
    if logging_info:
        print(f"{int(extract_time)//60}m{int(extract_time)%60}s")

    # * Generate location code for best level found
    if logging_info:
        print(">>> Generating address code: ", end='')
    start_time = time()
    best_lvl_cols = [f'best_level_{i}' for i in AVAIL_LEVELS]
    generated_data = parallelize_dataframe(
        extracted_data,
        generate_loc_code,
        n_cores=n_cores,
        best_lvl_cols=best_lvl_cols
    )
    code_gen_time = time() - start_time
    if logging_info:
        print(f"{int(code_gen_time)//60}m{int(code_gen_time)%60}s")

    # * Concat to original data
    final_data = pd.concat([generated_data, na_address_data])

    # * Concat with origin columns
    new_cols = [
        'level_1',
        'best_level_1',
        'level_1_code',
        'level_2',
        'best_level_2',
        'level_2_code',
        'level_3',
        'best_level_3',
        'level_3_code',
        'remained_address'
    ]
    final_data = pd.concat([
        data[orig_cols],
        final_data[new_cols]
    ], axis=1)

    return final_data



from multiprocessing import Pool

from tqdm import tqdm
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.models import load_model

from preprocessing_pgp.name.vector_creation import load_vectorization_from_disk


# Positional Embedding for Transformer
class PositionalEmbedding(layers.Layer):
    def __init__(self, sequence_length, input_dim, output_dim, **kwargs):
        super().__init__(**kwargs)
        self.token_embeddings = layers.Embedding(
            input_dim=input_dim, output_dim=output_dim)
        self.position_embeddings = layers.Embedding(
            input_dim=sequence_length, output_dim=output_dim)
        self.sequence_length = sequence_length
        self.input_dim = input_dim
        self.output_dim = output_dim

    def call(self, inputs):
        length = tf.shape(inputs)[-1]
        positions = tf.range(start=0, limit=length, delta=1)
        embedded_tokens = self.token_embeddings(inputs)
        embedded_positions = self.position_embeddings(positions)
        return embedded_tokens + embedded_positions

    def compute_mask(self, inputs, mask=None):
        return tf.math.not_equal(inputs, 0)

    def get_config(self):
        config = super(PositionalEmbedding, self).get_config()
        config.update({
            'output_dim': self.output_dim,
            'sequence_length': self.sequence_length,
            'input_dim': self.input_dim
        })

        return config


# Encoder model
class TransformerEncoder(layers.Layer):
    def __init__(self, embed_dim, dense_dim, num_heads, dropout=0.5, **kwargs):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.dense_dim = dense_dim
        self.num_heads = num_heads
        self.attention = layers.MultiHeadAttention(
            num_heads=num_heads, key_dim=embed_dim)
        if dropout > 0:
            self.dense_proj = tf.keras.Sequential(
                [layers.Dense(dense_dim, activation="relu"),
                 layers.Dropout(dropout),
                 layers.Dense(embed_dim), ]
            )
            self.dropout = layers.Dropout(dropout)
        else:
            self.dense_proj = tf.keras.Sequential(
                [layers.Dense(dense_dim, activation='relu'),
                 layers.Dense(embed_dim), ]
            )

        self.layernorm_1 = layers.LayerNormalization()
        self.layernorm_2 = layers.LayerNormalization()

    def call(self, inputs, mask=None):
        if mask is not None:
            mask = mask[:, tf.newaxis, :]
        attention_output = self.attention(
            inputs, inputs, attention_mask=mask)
        if hasattr(self, 'dropout') and self.dropout is not None:
            attention_output = self.dropout(attention_output)
        proj_input = self.layernorm_1(inputs + attention_output)
        proj_output = self.dense_proj(proj_input)
        return self.layernorm_2(proj_input + proj_output)

    def get_config(self):
        config = super().get_config()
        config.update({
            "embed_dim": self.embed_dim,
            "num_heads": self.num_heads,
            "dense_dim": self.dense_dim,
        })
        return config


# Decoder Model
class TransformerDecoder(layers.Layer):
    def __init__(self, embed_dim, dense_dim, num_heads, dropout=0.5, **kwargs):
        super().__init__(**kwargs)

        self.embed_dim = embed_dim
        self.dense_dim = dense_dim
        self.num_heads = num_heads
        self.attention_1 = layers.MultiHeadAttention(
            num_heads=num_heads, key_dim=embed_dim)
        self.attention_2 = layers.MultiHeadAttention(
            num_heads=num_heads, key_dim=embed_dim)
        if dropout > 0:
            self.dense_proj = tf.keras.Sequential(
                [layers.Dense(dense_dim, activation="relu"),
                 layers.Dropout(dropout),
                 layers.Dense(embed_dim), ]
            )
            self.dropout = layers.Dropout(dropout)
        else:
            self.dense_proj = tf.keras.Sequential(
                [layers.Dense(dense_dim, activation='relu'),
                 layers.Dense(embed_dim), ]
            )

        self.layernorm_1 = layers.LayerNormalization()
        self.layernorm_2 = layers.LayerNormalization()
        self.layernorm_3 = layers.LayerNormalization()
        self.supports_masking = True

    def get_config(self):
        config = super().get_config()
        config.update({
            'embed_dim': self.embed_dim,
            'num_heads': self.num_heads,
            'dense_dim': self.dense_dim
        })

        return config

    def get_causal_attention_mask(self, inputs):
        input_shape = tf.shape(inputs)
        batch_size, sequence_length = input_shape[0], input_shape[1]
        i = tf.range(sequence_length)[:, tf.newaxis]
        j = tf.range(sequence_length)
        mask = tf.cast(i >= j, dtype='int32')
        mask = tf.reshape(mask, (1, sequence_length, sequence_length))
        mult = tf.concat(
            [tf.expand_dims(batch_size, -1),
             tf.constant([1, 1], dtype=tf.int32)],
            axis=0
        )

        return tf.tile(mask, mult)

    def call(self, inputs, encoder_outputs, mask=None):
        causal_mask = self.get_causal_attention_mask(inputs)

        if mask is not None:
            padding_mask = tf.cast(mask[:, tf.newaxis, :], dtype='int32')
            padding_mask = tf.minimum(padding_mask, causal_mask)

        attention_output_1 = self.attention_1(
            query=inputs,
            value=inputs,
            key=inputs,
            attention_mask=causal_mask)
        if hasattr(self, 'dropout') and self.dropout is not None:
            attention_output_1 = self.dropout(attention_output_1)
        attention_output_1 = self.layernorm_1(inputs + attention_output_1)
        attention_output_2 = self.attention_2(
            query=attention_output_1,
            value=encoder_outputs,
            key=encoder_outputs,
            attention_mask=padding_mask,
        )
        if hasattr(self, 'dropout') and self.dropout is not None:
            attention_output_2 = self.dropout(attention_output_2)
        attention_output_2 = self.layernorm_2(
            attention_output_1 + attention_output_2)
        proj_output = self.dense_proj(attention_output_2)
        return self.layernorm_3(attention_output_2 + proj_output)


# Entire Transformer model
class TransformerModel:
    def __init__(self, source_vectorization, target_vectorization,
                 config_dict,
                 model_path=None, weight_path=None):

        self.source_vectorization = self.load_vectorization(
            source_vectorization)
        self.target_vectorization = self.load_vectorization(
            target_vectorization)

        self.load_model(model_path)
        self.load_model_weights(weight_path)

        self.sequence_length = config_dict['SEQUENCE_LENGTH']
        self.vocab_size = config_dict['VOCAB_SIZE']
        self.embed_dim = config_dict['EMBED_DIM']
        self.dense_dim = config_dict['DENSE_DIM']
        self.num_heads = config_dict['NUM_HEADS']
        self.drop_out = config_dict['DROPOUT_RATE']
        self.drop_out_enc = config_dict['DROPOUT_ENC']
        self.drop_out_dec = config_dict['DROPOUT_DEC']

    def load_vectorization(self, vectorization):
        if isinstance(vectorization, str):
            return load_vectorization_from_disk(vectorization)
        else:
            return vectorization

    def load_model(self, model_path: str = None):
        if model_path is None:
            return

        self.model = load_model(
            model_path,
            custom_objects={
                'PositionalEmbedding': PositionalEmbedding,
                'TransformerDecoder': TransformerDecoder,
                'TransformerEncoder': TransformerEncoder
            }
        )

    def load_model_weights(self, weight_path: str = None):
        if weight_path is None:
            return

        if not hasattr(self, 'model') or self.model is None:
            raise TypeError(
                'Please build the model or load a pre-trained model first')

        self.model.load_weights(weight_path)

    def build_model(self, *arg, **kwargs):
        encoder_inputs = tf.keras.Input(
            shape=(None,), dtype='int64', name='stripped')
        x = PositionalEmbedding(
            self.sequence_length, self.vocab_size, self.embed_dim)(encoder_inputs)
        encoder_outputs = TransformerEncoder(
            self.embed_dim, self.dense_dim, self.num_heads, self.drop_out_enc)(x)

        decoder_inputs = tf.keras.Input(
            shape=(None,), dtype='int64', name='original')
        x = PositionalEmbedding(
            self.sequence_length, self.vocab_size, self.embed_dim)(decoder_inputs)
        x = TransformerDecoder(
            self.embed_dim, self.dense_dim, self.num_heads, self.drop_out_dec)(x, encoder_outputs)

        if self.drop_out > 0:
            x = layers.Dropout(self.drop_out)(x)

        decoder_outputs = layers.Dense(
            self.vocab_size, activation='softmax')(x)

        self.model = tf.keras.Model(
            inputs=[encoder_inputs, decoder_inputs], outputs=[decoder_outputs])

        self.model.compile(*arg, **kwargs)

    def fit(self, *arg, **kwargs):
        if not hasattr(self, 'model') or self.model is None:
            raise TypeError(
                'Please build the model or load a pre-trained model first')

        return self.model.fit(*arg, **kwargs)

    def evaluate(self, *args, **kwargs):
        if not hasattr(self, 'model') or self.model is None:
            raise TypeError(
                'Please build the model or load a pre-trained model first')

        rs = self.model.evaluate(*args, **kwargs)
        print(rs)
        return rs

    def predict(self, input_sentence):
        if not hasattr(self, 'model') or self.model is None:
            raise TypeError(
                'Please build the model or load a pre-trained model first')

        if not hasattr(self, 'original_vocal'):
            self.original_vocal = self.target_vectorization.get_vocabulary()
            self.original_index_lookup = dict(
                zip(range(len(self.original_vocal)), self.original_vocal))

        input_word = input_sentence.split()
        tokenized_input_sentence = self.source_vectorization([input_sentence])
        decoded_sentence = '[start]'
        
        for i in range(min(self.sequence_length, len(input_word))):
            tokenized_target_sentence = self.target_vectorization([decoded_sentence])[
                :, :-1]
            predictions = self.model(
                [tokenized_input_sentence, tokenized_target_sentence]).numpy()
            
            # Change to beam search here if possible
            sampled_token_index = np.argmax(predictions[0, i, :])
            sampled_token = self.original_index_lookup[
                sampled_token_index] if sampled_token_index != 1 else input_word[i]
            decoded_sentence += ' ' + sampled_token

        return decoded_sentence.replace('[start]', '').strip()
    
    def predict_multi(self, sentences: pd.Series,
                      multiprocessing: bool = False,
                      n_cpu: int = 1) -> pd.Series:
        """
        Predict multiple sentences using whether parallel or single process

        Parameters
        ----------
        sentences : pd.Series
            The input sentence series containing the names
        multiprocessing : bool, optional
            Whether using parallel execution, by default False
        n_cpu : int, optional
            The number of cpu use in parallel execution, by default 1

        Returns
        -------
        pd.Series
            The returned prediction name series
        """
        if multiprocessing:
            with Pool(processes=n_cpu) as pool:
                pred_sentences = list(
                    tqdm(pool.imap(self.predict, sentences), total=sentences.size))

        else:
            pred_sentences = []
            for sent in tqdm(sentences):
                pred_sentences.append(self.predict(sent))

        pred_result = pd.Series(pred_sentences, dtype='str', name='pred_sent')

        return pred_result




"""
Module contains utility functions to help the email's functions
"""

import re
from typing import List, Tuple

import pandas as pd
import dns.resolver

from preprocessing_pgp.name.utils import is_name_accented


def is_valid_email_domain(
    domain: str
) -> bool:
    """
    Check email domain using resolver
    """
    try:
        dns.resolver.resolve(domain, 'MX')
        return True
    except:
        return False


def split_email(email: str) -> List[str]:
    """
    Split email into email's name & group (default by `@`)

    Parameters
    ----------
    email : str
        The original email

    Returns
    -------
    List[str]
        The list contains email's `name` and `group`
    """
    if not email:
        return [None, None]

    split_result = email.split('@', maxsplit=1)

    if len(split_result) == 2:
        return split_result

    return [None, None]


def clean_email_name(
    data: pd.DataFrame,
    name_col: str = 'email_name'
) -> pd.DataFrame:
    """
    Process cleaning email name data -- removing '.' and other processing

    Parameters
    ----------
    data : pd.DataFrame
        The original dataframe
    name_col : str, optional
        The column name contains email_name, by default 'email_name'

    Returns
    -------
    str
        Cleaned email name data with additional column `cleaned_email_name`
    """
    data[f'cleaned_{name_col}'] = ''

    # * Removing accented name
    data.loc[
        data[name_col].apply(is_name_accented),
        f'cleaned_{name_col}'
    ] = None

    # * Removing names with spaces
    data.loc[
        data[name_col].str.contains(r'\s+', regex=True),
        f'cleaned_{name_col}'
    ] = None

    # * Remove digit, '.' from name & lower name
    clean_data_mask = data[f'cleaned_{name_col}'].notna()
    data.loc[
        clean_data_mask,
        f'cleaned_{name_col}'
    ] = data.loc[
        clean_data_mask,
        name_col
    ].str.replace(r'\d+', '', regex=True)\
        .str.replace('.', '', regex=True)\
        .str.lower()

    # * Filter other case non-cleaned
    data.loc[
        data[f'cleaned_{name_col}'] == '',
        f'cleaned_{name_col}'
    ] = None

    return data


def sort_series_by_appearance(
    series: pd.Series
) -> pd.Series:
    """
    Sort series by the order of appearance in descending order

    Parameters
    ----------
    series : pd.Series
        The input series to sort for appearance

    Returns
    -------
    pd.Series
        The sorted descending series
    """

    name = series.name

    sorted_series = series.copy()\
        .value_counts().reset_index(name='count')\
        .rename(columns={'index': name})[name]

    return sorted_series


def extract_sub_string(
    sub_regex: str,
    string: str,
    take_first: bool = True
) -> Tuple[str, str]:
    """
    Extract substring from string given the regex

    Then return the sub_string and the remaining string

    If not found return empty string `''`

    Parameters
    ----------
    sub_regex : str
        The regex to extract from string
    string : str
        The input string to extract from
    take_first : bool
        Whether to take the first appearance, by default True

    Returns
    -------
    Tuple[str, str]
        `sub_string` and `remained string`
    """
    found_subs = re.findall(sub_regex, string)

    # * Not found any substring match
    if len(found_subs) == 0:
        return ('', string)

    # * Found at least one
    if take_first:
        sub_string = found_subs[0]
    else:
        sub_string = found_subs[-1]
    remain_string = re.sub(sub_string, '', string)

    return sub_string, remain_string



"""
Module contains tools for processing name
"""

import pandas as pd
from unidecode import unidecode

from preprocessing_pgp.name.const import NICKNAME_REGEX


def remove_nicknames(
    name_df: pd.DataFrame,
    name_col: str
) -> pd.DataFrame:
    """
    Remove nicknames in name column given the data

    Parameters
    ----------
    name_df : pd.DataFrame
        The data process
    name_col : str
        The name column in data to remove nickname

    Returns
    -------
    pd.DataFrame
        Cleaned data without nickname -- added new column `clean_name_col`
    """
    name_df[name_col] =\
        name_df[name_col].str.replace(NICKNAME_REGEX, '', regex=True)\
        .str.strip()

    return name_df


def is_name_accented(name: str) -> bool:
    """
    Check whether the name is accented or not

    Parameters
    ----------
    name : str
        The input name

    Returns
    -------
    bool
        Whether the name is accented
    """
    return unidecode(name) != name



import json
from time import time
from typing import Tuple
import warnings
import logging

import pandas as pd
from tensorflow import keras
from tqdm import tqdm

from preprocessing_pgp.name.name_processing import NameProcessor
from preprocessing_pgp.name.model.transformers import TransformerModel
from preprocessing_pgp.name.type.extractor import process_extract_name_type
from preprocessing_pgp.name.const import (
    MODEL_PATH,
    RULE_BASED_PATH
)
from preprocessing_pgp.utils import (
    sep_display,
    parallelize_dataframe,
)

tqdm.pandas()
warnings.filterwarnings("ignore")
logger = logging.getLogger()
logger.setLevel(logging.CRITICAL)


class EnrichName:
    """
    Wrap-up module to enrich and filling accent to names
    """

    def __init__(
        self,
        model_weight_path: str,
        vectorization_paths: Tuple[str, str],
        model_config_path: str,
        name_rb_pth: str
    ) -> None:
        start_time = time()
        self.model = self.load_model(
            model_weight_path,
            vectorization_paths,
            model_config_path
        )
        self.fname_rb = f'{name_rb_pth}/firstname_dict.parquet'
        self.mname_rb = f'{name_rb_pth}/middlename_dict.parquet'
        self.lname_rb = f'{name_rb_pth}/lastname_dict.parquet'
        self.name_processor = NameProcessor(
            self.model,
            self.fname_rb,
            self.mname_rb,
            self.lname_rb,
        )
        # Timing
        self.total_load_time = time() - start_time

    def load_model(
        self,
        model_weight_path: str,
        vectorization_paths: Tuple[str, str],
        model_config_path: str
    ) -> TransformerModel:
        start_time = time()
        # ? Load config dict
        with open(model_config_path) as json_file:
            config_dict = json.load(json_file)

        # ? BUILD & LOAD MODEL
        source_vec_pth, target_vec_pth = vectorization_paths
        transformer = TransformerModel(
            source_vectorization=source_vec_pth,
            target_vectorization=target_vec_pth,
            config_dict=config_dict
        )
        transformer.build_model(
            optimizer=keras.optimizers.Adam(
                learning_rate=config_dict['LEARNING_RATE']),
            loss='sparse_categorical_crossentropy',
            metrics=['accuracy']
        )
        transformer.load_model_weights(model_weight_path)

        self.model_load_time = time() - start_time

        return transformer

    def refill_accent(
        self,
        name_df: pd.DataFrame,
        name_col: str
    ) -> pd.DataFrame:
        return self.name_processor.fill_accent(name_df, name_col)

    def get_time_report(self) -> pd.DataFrame:
        return pd.DataFrame({
            'model load time': [self.model_load_time],
            'total load time': [self.total_load_time],
        })


def enrich_clean_data(
    clean_df: pd.DataFrame,
    name_col: str,
) -> pd.DataFrame:
    """
    Applying the model of filling accent to cleaned Vietnamese names

    Parameters
    ----------
    clean_df : pd.DataFrame
        The dataframe containing cleaned names
    name_col : str
        The column name that holds the raw names

    Returns
    -------
    pd.DataFrame
        The final dataframe that contains:

        * `name_col`: raw names -- input name column
        * `predict`: predicted names using model only
        * `final`: beautified version of prediction with additional rule-based approach
    """
    if clean_df.empty:
        return clean_df
    model_weight_path = f'{MODEL_PATH}/best_transformer_model.h5'
    vectorization_paths = (
        f'{MODEL_PATH}/vecs/source_vectorization_layer.pkl',
        f'{MODEL_PATH}/vecs/target_vectorization_layer.pkl'
    )
    model_config_path = f'{MODEL_PATH}/hp.json'

    enricher = EnrichName(
        model_weight_path=model_weight_path,
        vectorization_paths=vectorization_paths,
        model_config_path=model_config_path,
        name_rb_pth=RULE_BASED_PATH
    )

    final_df = enricher.refill_accent(
        clean_df,
        name_col
    )

    return final_df


def process_enrich(
    data: pd.DataFrame,
    name_col: str = 'name',
    n_cores: int = 1,
    logging_info: bool = True
) -> pd.DataFrame:
    """
    Applying the model of filling accent to non-accent Vietnamese names
    received from the `raw_df` at `name_col` column.

    Parameters
    ----------
    data : pd.DataFrame
        The dataframe containing raw names
    name_col : str
        The column name that holds the raw names, by default 'name'
    n_cores : int
        The number of cores used to run parallel, by default 1 core is used
    logging_info : bool
        Whether to log info about run time, by default True

    Returns
    -------
    pd.DataFrame
        The final dataframe that contains:

        * `customer_type`: the type of customer extracted from name
        * `predict`: predicted names using model only
        * `final`: beautified version of prediction with additional rule-based approach
    """
    orig_cols = data.columns

    # * Na names & filter out name col
    na_data = data[data[name_col].isna()][[name_col]]
    cleaned_data = data[data[name_col].notna()][[name_col]]

    # * Extracting customer type -- Only enrich 'customer' type
    cleaned_data = process_extract_name_type(
        cleaned_data,
        name_col,
        n_cores=n_cores,
        logging_info=logging_info
    )
    sep_display()
    customer_data = cleaned_data.query('customer_type == "customer"')
    #                    | (cleaned_data[name_col].str.split(' ').str.len() > 3))
    non_customer_data = cleaned_data.query('customer_type != "customer"')

    # # Clean names -- Not Needed
    # start_time = time()
    # if n_cores == 1:
    #     customer_data = preprocess_df(
    #         customer_data,
    #         name_col=name_col
    #     )
    # else:
    #     customer_data = parallelize_dataframe(
    #         customer_data,
    #         preprocess_df,
    #         n_cores=n_cores,
    #         name_col=name_col
    #     )
    # clean_time = time() - start_time
    # print(f"Cleansing takes {int(clean_time)//60}m{int(clean_time)%60}s")
    # sep_display()

    # Enrich names
    if logging_info:
        print(">>> Enriching names: ", end='')
    start_time = time()
    enriched_data = parallelize_dataframe(
        customer_data,
        enrich_clean_data,
        n_cores=n_cores,
        name_col=name_col
    )
    enrich_time = time() - start_time
    if logging_info:
        print(f"{int(enrich_time)//60}m{int(enrich_time)%60}s")

    # * Concat na data
    new_cols = [
        'customer_type',
        'predict',
        'final',
        'last_name',
        'middle_name',
        'first_name',
    ]
    na_data[new_cols] = None
    final_data = pd.concat([
        enriched_data,
        non_customer_data,
        na_data
    ])

    # * Concat with original cols
    final_data = pd.concat([data[orig_cols], final_data[new_cols]], axis=1)

    return final_data



"""
Module contains objects and functions to support extracting information from email
"""

from time import time

import pandas as pd

from preprocessing_pgp.utils import parallelize_dataframe, sep_display
from preprocessing_pgp.email.validator import process_validate_email
from preprocessing_pgp.email.extractors.email_name_extractor import EmailNameExtractor
from preprocessing_pgp.email.extractors.email_yob_extractor import EmailYOBExtractor
from preprocessing_pgp.email.extractors.email_phone_extractor import EmailPhoneExtractor
from preprocessing_pgp.email.extractors.email_address_extractor import EmailAddressExtractor


class EmailInfoExtractor:
    """
    Class contains function to extract for email information
    """

    def __init__(self):
        # Using factory pattern for better view
        self.name_extractor = EmailNameExtractor()
        self.yob_extractor = EmailYOBExtractor()
        self.phone_extractor = EmailPhoneExtractor()
        self.address_extractor = EmailAddressExtractor()

    def extract_info(
        self,
        data: pd.DataFrame,
        email_name_col: str = 'email_name'
    ) -> pd.DataFrame:
        """
        Extract any information from email's name if possible

        Parameters
        ----------
        data : pd.DataFrame
            The input data contains an email_name column
        email_name_col : str, optional
            The name of the column contains email's name, by default 'email_name'

        Returns
        -------
        pd.DataFrame
            Data with additional info columns:
            * `username_extracted` : Extracted username from email name
            * `gender_extracted` : Extracted gender from username
            * `yob_extracted` : Extracted year of birth from email name
            * `phone_extracted` : Extracted phone number from email name
            * `address_extracted` : Extracted address from email name
        """
        if data.empty:
            return data
        # * Extracting name & gender
        extracted_data =\
            self.name_extractor.extract_username(data, email_name_col)

        # * Extracting yob
        extracted_data =\
            self.yob_extractor.extract_yob(extracted_data, email_name_col)

        # * Extracting phone
        extracted_data =\
            self.phone_extractor.extract_phone(extracted_data, email_name_col)

        # * Extracting address
        extracted_data =\
            self.address_extractor.extract_address(
                extracted_data, email_name_col)

        return extracted_data


def process_extract_email_info(
    data: pd.DataFrame,
    email_col: str = 'email',
    n_cores: int = 1
) -> pd.DataFrame:
    """
    Process extracting information from email, extracted information may conclude:
    1. Username -- with accent
    2. Customer type -- derived from username
    3. Year of birth
    4. Address -- 3 levels
    5. Email group
    6. Auto-email

    Parameters
    ----------
    data : pd.DataFrame
        The input data contains an email column
    email_col : str
        The name of the column contains email's records, by default `email`
    n_cores

    Returns
    -------
    pd.DataFrame
        Original data with additional columns contains 6 additional information as above
    """

    # * Select the part of data to process
    email_data = data[[email_col]]
    orig_cols = data.columns

    info_extractor = EmailInfoExtractor()
    # ? Validate email -- Only extract info for valid email
    validated_data = process_validate_email(
        email_data,
        email_col=email_col,
        n_cores=n_cores
    )
    sep_display()
    valid_email = validated_data.query('is_email_valid')
    invalid_email = validated_data.query('~is_email_valid')

    if valid_email.empty:
        return invalid_email
    # ? Separate email name and group
    valid_email[f'{email_col}_name'] =\
        valid_email[f'cleaned_{email_col}'].str.split('@').str[0]

    # ? Extract username from email
    start_time = time()
    extracted_valid_email = parallelize_dataframe(
        valid_email,
        info_extractor.extract_info,
        n_cores=n_cores,
        email_name_col=f'{email_col}_name'
    )
    extract_time = time() - start_time
    print(">>> Extracting information from email: ", end='')
    print(f"{int(extract_time)//60}m{int(extract_time)%60}s")

    final_data = pd.concat([extracted_valid_email, invalid_email])

    # * Generate whether username is certain
    final_data['username_iscertain'] =\
        (final_data['is_email_valid'])\
        & (final_data['customer_type'] == 'customer')

    # * Concat with data of other columns
    extracted_cols = [
        f'{email_col}_name',
        f'cleaned_{email_col}',
        'is_email_valid',
        'is_autoemail',
        'username_iscertain',
        'email_domain',
        'private_email',
        'customer_type',
        'username_extracted',
        'enrich_name',
        'gender_extracted',
        'yob_extracted',
        'phone_extracted',
        'address_extracted',
    ]
    final_data = pd.concat(
        [data[orig_cols], final_data[extracted_cols]], axis=1)

    return final_data



from abc import abstractmethod

from preprocessing_pgp.phone.const import (
    SUB_PHONE_10NUM,
    SUB_PHONE_11NUM,
    SUB_TELEPHONE_10NUM,
    SUB_TELEPHONE_11NUM,
    PHONE_LENGTH
)


class PhoneValidator:
    """
    Abstract Class contains validating functions for verify phone
    """

    @abstractmethod
    def is_valid_phone(self, phone: str) -> bool:
        """
        Check whether the phone is valid

        Parameters
        ----------
        phone : str
            The input phone string to check for validation

        Returns
        -------
        bool
            Whether it is valid or not

        Raises
        ------
        NotImplementedError
            The subclass not override this function
        """
        raise NotImplementedError("Subclasses should implement this function!")


class MobiPhoneValidator(PhoneValidator):
    """
    Class to check for mobi phone syntax
    """

    def __init__(self):
        self.old_length = PHONE_LENGTH['old_mobi']
        self.new_length = PHONE_LENGTH['new_mobi']

    def _is_new_sub_phone(self, phone: str) -> bool:
        return phone[:3] in SUB_PHONE_10NUM

    def _is_old_sub_phone(self, phone: str) -> bool:
        return phone[:3] in SUB_PHONE_11NUM

    def _is_new_phone_length(self, phone: str) -> bool:
        return len(phone) == self.new_length

    def _is_old_phone_length(self, phone: str) -> bool:
        return len(phone) == self.old_length

    def is_new_phone(self, phone: str) -> bool:
        """
        Check whether the phone is `new mobi` phone

        Parameters
        ----------
        phone : str
            The input phone number sequence

        Returns
        -------
        bool
            Whether the phone is `new mobi` phone
        """

        return self._is_new_phone_length(phone)\
            and self._is_new_sub_phone(phone)

    def is_old_phone(self, phone: str) -> bool:
        """
        Check whether the phone is `old mobi` phone

        Parameters
        ----------
        phone : str
            The input phone number sequence

        Returns
        -------
        bool
            Whether the phone is `old mobi` phone
        """

        return self._is_old_phone_length(phone)\
            and self._is_old_sub_phone(phone)

    def is_valid_phone(self, phone: str) -> bool:
        """
        General function to check whether the phone is valid phone number or not

        Parameters
        ----------
        phone : str
            The input phone number sequence

        Returns
        -------
        bool
            Whether the phone is valid phone
        """
        return self.is_new_phone(phone)\
            or self.is_old_phone(phone)


class LandlinePhoneValidator(PhoneValidator):
    """
    Class to check for mobi phone syntax
    """

    def __init__(self):
        self.old_length = PHONE_LENGTH['old_landline']
        self.new_length = PHONE_LENGTH['new_landline']

    def _is_new_sub_phone(self, phone: str) -> bool:
        return phone[:3] in SUB_TELEPHONE_11NUM\
            or phone[:4] in SUB_TELEPHONE_11NUM

    def _is_old_sub_phone(self, phone: str) -> bool:
        return phone[:2] in SUB_TELEPHONE_10NUM\
            or phone[:3] in SUB_TELEPHONE_10NUM\
            or phone[:4] in SUB_TELEPHONE_10NUM

    def _is_new_phone_length(self, phone: str) -> bool:
        return len(phone) == self.new_length

    def _is_old_phone_length(self, phone: str) -> bool:
        return len(phone) == self.old_length

    def is_new_phone(self, phone: str) -> bool:
        """
        Check whether the phone is `new mobi` phone

        Parameters
        ----------
        phone : str
            The input phone number sequence

        Returns
        -------
        bool
            Whether the phone is `new mobi` phone
        """

        return self._is_new_phone_length(phone)\
            and self._is_new_sub_phone(phone)

    def is_old_phone(self, phone: str) -> bool:
        """
        Check whether the phone is `old mobi` phone

        Parameters
        ----------
        phone : str
            The input phone number sequence

        Returns
        -------
        bool
            Whether the phone is `old mobi` phone
        """

        return self._is_old_phone_length(phone)\
            and self._is_old_sub_phone(phone)

    def is_valid_phone(self, phone: str) -> bool:
        """
        General function to check whether the phone is valid phone number or not

        Parameters
        ----------
        phone : str
            The input phone number sequence

        Returns
        -------
        bool
            Whether the phone is valid phone
        """
        return self.is_new_phone(phone)\
            or self.is_old_phone(phone)
